class Config:
    def __init__(self):
        self.class_names = ['person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat',
                       'traffic light', 'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog',
                       'horse', 'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag',
                       'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
                       'baseball glove', 'skateboard', 'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup', 'fork',
                       'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog',
                       'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv',
                       'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
                       'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush']
        self.model_weight = None
        self.device = 'cpu'
        self.img_size = (320, 320)
        self.confidence = 0.4
        self.iou_thresh = 0.45



class Config:
    def __init__(self):
        self.device = 'cpu'
        self.min_face_size = 20.0
        self.thresholds = (0.6, 0.7, 0.8)
        self.min_detection_size = 12
        self.factor = 0.707
        self.top_k = 5000
        self.nms_thresholds = 0.4
        self.keep_top_k = 750
        self.visualization_threshold = 0.6
        self.resnet50_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/resnet50.pth"
        self.resnet50_cache = 'weights/vision/face_detection/retinaface/torch/resnet50.pth'
        self.resnet50 = None
        self.mobilenet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/mobilenet.pth"
        self.mobilenet_cache = 'weights/vision/face_detection/retinaface/torch/mobilenet.pth'
        self.mobilenet = None
        self.mobilenetV1X_url = 'https://github.com/Practical-AI/deep_utils/releases/download/0.2.0/mobilenetV1X0.25_pretrain.tar'
        self.mobilenetV1X_cache = 'weights/vision/face_detection/retinaface/torch/mobilenetV1X0.25_pretrain.tar'
        self.mobilenetV1X = None
        self.network = 'mobilenet'
        self.cfg_mnet = {
            'name': 'mobilenet',
            'min_sizes': [[16, 32], [64, 128], [256, 512]],
            'steps': [8, 16, 32],
            'variance': [0.1, 0.2],
            'clip': False,
            'loc_weight': 2.0,
            'gpu_train': True,
            'batch_size': 32,
            'ngpu': 1,
            'epoch': 250,
            'decay1': 190,
            'decay2': 220,
            'image_size': 640,
            'pretrain': True,
            'return_layers': {'stage1': 1, 'stage2': 2, 'stage3': 3},
            'in_channel': 32,
            'out_channel': 64
        }

        self.cfg_re50 = {
            'name': 'resnet50',
            'min_sizes': [[16, 32], [64, 128], [256, 512]],
            'steps': [8, 16, 32],
            'variance': [0.1, 0.2],
            'clip': False,
            'loc_weight': 2.0,
            'gpu_train': True,
            'batch_size': 24,
            'ngpu': 4,
            'epoch': 100,
            'decay1': 70,
            'decay2': 90,
            'image_size': 840,
            'pretrain': True,
            'return_layers': {'layer2': 1, 'layer3': 2, 'layer4': 3},
            'in_channel': 256,
            'out_channel': 256
        }



def show_plt(img, is_rgb, figsize=None):
    import matplotlib
    matplotlib.use('TkAgg')
    import matplotlib.pyplot as plt
    if not is_rgb:
        img = img[..., ::-1]
    plt.figure(figsize=figsize)
    plt.imshow(img)
    plt.show()



class AugmentTorch:

    @staticmethod
    def random_resized_crop(size, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.),
                            interpolation="BILINEAR"):
        from torchvision import transforms
        if interpolation == "BILINEAR":
            from torchvision.transforms import InterpolationMode
            interpolation = InterpolationMode.BILINEAR
        return transforms.RandomResizedCrop(size, scale=scale, ratio=ratio, interpolation=interpolation)

    @staticmethod
    def five_crop(size):
        from torchvision import transforms
        return transforms.FiveCrop(size)

    @staticmethod
    def ten_crop(size):
        from torchvision import transforms
        return transforms.TenCrop(size)

    @staticmethod
    def random_horizontal_flip(p=0.5):
        from torchvision import transforms
        return transforms.RandomHorizontalFlip(p=p)

    @staticmethod
    def random_affine(degrees, translate=None, scale=None, shear=None, interpolation="NEAREST", fill=0,
                      fillcolor=None, resample=None):
        from torchvision import transforms
        if interpolation == "NEAREST":
            from torchvision.transforms import InterpolationMode
            interpolation = InterpolationMode.NEAREST
        return transforms.RandomAffine(degrees, translate=translate, scale=scale, shear=shear,
                                       interpolation=interpolation, fill=fill,
                                       fillcolor=fillcolor, resample=resample)

    @staticmethod
    def random_rotation(degrees, interpolation="NEAREST",
                        expand=False, center=None, fill=0, resample=None):
        from torchvision import transforms
        if interpolation == "NEAREST":
            from torchvision.transforms import InterpolationMode
            interpolation = InterpolationMode.NEAREST
        return transforms.RandomRotation(degrees, interpolation=interpolation, expand=expand, center=center, fill=fill,
                                         resample=resample)

    @staticmethod
    def gaussian_blur(kernel_size, sigma=(0.1, 2.0)):
        from torchvision import transforms
        return transforms.GaussianBlur(kernel_size=kernel_size, sigma=sigma)

    @staticmethod
    def random_erasing(p=0.5, scale=(0.02, 0.33), ratio=(0.3, 3.3), value=0, inplace=False):
        from torchvision import transforms
        return transforms.RandomErasing(p=p, scale=scale, ratio=ratio, value=value, inplace=inplace)

    @staticmethod
    def center_crop(size):
        from torchvision import transforms
        return transforms.CenterCrop(size)

    @staticmethod
    def color_jitter(brightness=0, contrast=0, saturation=0, hue=0):
        from torchvision import transforms
        return transforms.ColorJitter(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue)

    @staticmethod
    def random_vertical_flip(p=0.5):
        from torchvision import transforms
        return transforms.RandomVerticalFlip(p=p)

    @staticmethod
    def pad(padding, fill=0, padding_mode="constant"):
        from torchvision import transforms
        return transforms.Pad(padding=padding, fill=fill, padding_mode=padding_mode)

    @staticmethod
    def random_crop(size, padding=None, pad_if_needed=False, fill=0, padding_mode="constant"):
        from torchvision import transforms
        return transforms.RandomCrop(size=size, padding=padding, pad_if_needed=pad_if_needed, fill=fill,
                                     padding_mode=padding_mode)

    @staticmethod
    def random_perspective(distortion_scale=0.5, p=0.5, interpolation="BILINEAR", fill=0):
        from torchvision import transforms
        if interpolation == "BILINEAR":
            from torchvision.transforms import InterpolationMode
            interpolation = InterpolationMode.BILINEAR
        return transforms.RandomPerspective(distortion_scale=distortion_scale, p=p, interpolation=interpolation,
                                            fill=fill)

    @staticmethod
    def resize(size, interpolation="BILINEAR"):
        from torchvision import transforms
        if interpolation == "BILINEAR":
            from torchvision.transforms import InterpolationMode
            interpolation = InterpolationMode.BILINEAR
        return transforms.Resize(size=size, interpolation=interpolation)

    @staticmethod
    def normalize(mean, std, inplace=False):
        from torchvision import transforms
        return transforms.Normalize(mean=mean, std=std, inplace=inplace)

    @staticmethod
    def to_tensor():
        from torchvision import transforms
        return transforms.ToTensor()

    @staticmethod
    def get_augments(*args):
        from torchvision import transforms
        import torch
        transformations = []
        use_lambda = False
        for arg in args:
            prob = 0
            arg_type = type(arg)
            if arg_type is tuple:
                arg, prob = arg
            if type(arg) is callable:
                transform = arg()
            else:
                transform = arg
            if prob:
                transform = transforms.RandomApply([transform], p=prob)
            if type(transform) in [transforms.TenCrop, transforms.FiveCrop]:
                use_lambda = True
            if type(transform) in [transforms.Normalize]:
                if use_lambda:
                    to_append = transforms.Lambda(
                        lambda tensors: torch.stack([AugmentTorch.to_tensor()(t) for t in tensors]))
                else:
                    to_append = AugmentTorch.to_tensor()
                transformations.append(to_append)
            if use_lambda:
                transform = transforms.Lambda(lambda tensors: torch.stack([transform(t) for t in tensors]))
            transformations.append(transform)
        return transforms.Compose(transformations)


def tensor_to_image(tensor, mean=None, std=None, return_array=True):
    import torch
    from torchvision import transforms
    import numpy as np

    if mean is not None or std is not None:
        c = tensor.size()[0]
        if std is None:
            std = torch.ones(c)
        elif mean is None:
            mean = torch.zeros(c)

        tensor = transforms.Normalize(mean=[-m / s for m, s in zip(mean, std)], std=[1 / s for s in std])(tensor)

    image = transforms.ToPILImage()(tensor)
    if return_array:
        image = np.array(image)
    return image



from .main import *
from .ssd import *
from .mtcnn import *
from .haarcascade import *
from .ultralight import *
from .retinaface import *
from deep_utils.utils.lib_utils.main_utils import list_utils, loader

Face_Detection_Models = {
    "SSDCV2CaffeFaceDetector": SSDCV2CaffeFaceDetector,
    "MTCNNTorchFaceDetector": MTCNNTorchFaceDetector,
    "HaarcascadeCV2FaceDetector": HaarcascadeCV2FaceDetector,
    "MTCNNTFFaceDetector": MTCNNTFFaceDetector,
    "UltralightTorchFaceDetector": UltralightTorchFaceDetector,
    "UltralightTFFaceDetector": UltralightTFFaceDetector,
    "RetinaFaceTorchFaceDetector": RetinaFaceTorchFaceDetector,
}

list_face_detection_models = list_utils(Face_Detection_Models)


def face_detector_loader(name, **kwargs) -> FaceDetector:
    return loader(Face_Detection_Models, list_face_detection_models)(name, **kwargs)



from abc import abstractmethod
from deep_utils.main_abs.main import MainClass
from deep_utils.utils.utils.main import dictnamedtuple

OUTPUT_CLASS = dictnamedtuple("FaceDetector", ["boxes", "landmarks", "confidences"])


class FaceDetector(MainClass):
    def __init__(self, name, file_path, **kwargs):
        super().__init__(name, file_path=file_path, **kwargs)
        self.output_class = OUTPUT_CLASS

    @abstractmethod
    def detect_faces(self, img, is_rgb, confidence=None, get_time=False) -> OUTPUT_CLASS:
        pass



from deep_utils.utils.utils.main import shift_lst
from collections import OrderedDict


class BlocksTorch:
    @staticmethod
    def conv_norm_act(in_c, out_c, k=(3, 3), s=(1, 1), p=(1, 1),
                      norm='bn', act='relu', conv=True, index=0, conv_kwargs: dict = None,
                      norm_kwargs: dict = None, act_kwargs: dict = None,
                      pooling=None, pooling_k=(2, 2), pooling_s=(2, 2),
                      pool_kwargs=None, move_forward=0):
        from torch import nn
        conv_kwargs = dict() if conv_kwargs is None else conv_kwargs
        norm_kwargs = dict() if norm_kwargs is None else norm_kwargs
        act_kwargs = dict() if act_kwargs is None else act_kwargs
        pool_kwargs = dict() if pool_kwargs is None else pool_kwargs
        k = (k, k) if type(k) is int else k
        s = (s, s) if type(s) is int else s
        p = (p, p) if type(p) is int else p
        pooling_s = (pooling_s, pooling_s) if type(pooling_s) is int else pooling_s
        pooling_k = (pooling_k, pooling_k) if type(pooling_k) is int else pooling_k
        modules, names = [], []
        if conv:
            modules.append(BlocksTorch.conv_2d(in_c, out_c, k, s, p, **conv_kwargs))
            names.append(f"conv2d_{index}_in{in_c}_f{out_c}_k{k[0]}_s{s[0]}")
        if norm:
            modules.append(BlocksTorch.load_layer_norm(norm, out_c=out_c, **norm_kwargs))
            names.append(f"{norm}2d_{index}")
        if act:
            modules.append(BlocksTorch.load_activation(act, **act_kwargs))
            names.append(f"{act}_{index}")
        modules = shift_lst(modules, move_forward=move_forward % 3)
        names = shift_lst(names, move_forward=move_forward % 3)
        if pooling:
            pooling_layer = BlocksTorch.load_pooling(pooling, pooling_k, pooling_s, **pool_kwargs)
            modules.append(pooling_layer)
            names.append(f"{pooling}2d_{index}")
        cnn = nn.Sequential(OrderedDict({name: module for name, module in zip(names, modules)}))
        return cnn

    @staticmethod
    def load_pooling(pooling, k, s, **pooling_kwargs):
        from torch import nn
        if pooling == 'max':
            pooling = nn.MaxPool2d(k, s, **pooling_kwargs)
        elif pooling == 'avg':
            pooling = nn.AvgPool2d(k, s, **pooling_kwargs)
        else:
            raise ValueError(f"requested pooling {pooling} is not supported. Supported poolings: max, avg")
        return pooling

    @staticmethod
    def load_activation(act, **act_kwargs):
        from torch import nn
        if act == 'relu':
            activation = nn.ReLU(**act_kwargs)
        elif act == 'leaky_relu':
            activation = nn.LeakyReLU(**act_kwargs)
        else:
            raise ValueError(f"requested activation {act} is not supported. Supported activations: relu, leaky_relu")
        return activation

    @staticmethod
    def conv_2d(in_c, out_c, k=(3, 3), s=(1, 1), p=(1, 1), groups=1, bias=True, padding_mode='zeros'):
        import torch
        return torch.nn.Conv2d(in_c, out_c, k, s, p, groups=groups, bias=bias, padding_mode=padding_mode)

    @staticmethod
    def load_layer_norm(norm, out_c, **norm_kwargs):
        from torch import nn
        if norm == 'bn':
            layer_norm = nn.BatchNorm2d(num_features=out_c, **norm_kwargs)
        elif norm == 'gn':
            layer_norm = nn.GroupNorm(num_channels=out_c, **norm_kwargs)
        else:
            raise ValueError(
                f"requested layer_norm {norm} is not supported. Supported layer_norms: bn(batch normalization), "
                f"gn(group normalization")
        return layer_norm

    @staticmethod
    def res_basic_block(in_c, out_c, bias, down_sample=False):
        from torch import nn
        modules = []
        names = []
        for i in range(2):
            modules.append(
                BlocksTorch.conv_norm_act(int(in_c / 2) if down_sample and i == 0 else in_c, out_c,
                                          act=False if i == 1 else 'relu',
                                          s=(2, 2) if down_sample and i == 0 else (1, 1), conv_kwargs={'bias': bias}))
            names.append(f"cnn_{i}_in{in_c}_f{out_c}")
        res_block = nn.Sequential(OrderedDict({name: module for name, module in zip(names, modules)}))
        return res_block


if __name__ == '__main__':
    a = BlocksTorch.conv_norm_act(32, 64, move_forward=2)
    print(a)



import numpy as np
from deep_utils.utils.box_utils.boxes import Point


class VideoWriterCV:
    def __init__(self, save_path, width, height, fourcc="XVID", fps=30, colorful=True, in_source='Numpy'):
        import cv2
        point = Point.point2point((width, height), in_source=in_source, to_source=Point.PointSource.CV)
        fourcc = cv2.VideoWriter_fourcc(*fourcc) if type(fourcc) is str else fourcc
        self.vw = cv2.VideoWriter(save_path, fourcc, fps, point, colorful)

    def write(self, frame):
        self.vw.write(frame)


def rotate(img, rotation_degree, center_point=None, scale=1.0, dsize=None, bound=False, clockwise=True):
    import cv2
    h, w = img.shape[:2]
    (w, h) = dsize = (w, h) if dsize is None else dsize
    center_point = (w // 2, h // 2) if center_point is None else center_point
    # negative angle >> clockwise rotation | positive angle >> counter clockwise rotation
    rotation_degree = -rotation_degree if clockwise else rotation_degree
    m = cv2.getRotationMatrix2D(center_point, rotation_degree, scale)
    if bound:
        h, w = img.shape[:2]
        cos = abs(m[0, 0])
        sin = abs(m[0, 1])
        w_ = int((cos * w) + (sin * h))
        h_ = int((cos * h) + (sin * w))
        m[0, 2] += w_ // 2 - w // 2
        m[1, 2] += h_ // 2 - h // 2
        dsize = (w_, h_)
    rotated = cv2.warpAffine(img, m, dsize)

    return rotated


def translate(img, tx, ty, dsize=None):
    import cv2
    h, w = img.shape[:2][::-1]
    dsize = (w, h) if dsize is None else dsize
    translation_matrix = np.array([
        [1, 0, tx],
        [0, 1, ty]], dtype=np.float32)
    translated_image = cv2.warpAffine(src=img, M=translation_matrix, dsize=dsize)
    return translated_image


def show_destroy_cv2(img, win_name=''):
    import cv2
    try:
        cv2.imshow(win_name, img)
        cv2.waitKey(0)
        cv2.destroyWindow(win_name)
    except Exception as e:
        cv2.destroyWindow(win_name)
        raise e



# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ross Girshick
# --------------------------------------------------------

import time


class Timer(object):
    """A simple timer."""
    def __init__(self):
        self.total_time = 0.
        self.calls = 0
        self.start_time = 0.
        self.diff = 0.
        self.average_time = 0.

    def tic(self):
        # using time.time instead of time.clock because time time.clock
        # does not normalize for multithreading
        self.start_time = time.time()

    def toc(self, average=True):
        self.diff = time.time() - self.start_time
        self.total_time += self.diff
        self.calls += 1
        self.average_time = self.total_time / self.calls
        if average:
            return self.average_time
        else:
            return self.diff

    def clear(self):
        self.total_time = 0.
        self.calls = 0
        self.start_time = 0.
        self.diff = 0.
        self.average_time = 0.



import time
from functools import wraps


def get_method_time(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        tic = time.time()
        results = func(self, *args, **kwargs)
        toc = time.time()
        elapsed_time = round(toc - tic, 4)
        print(f"elapsed time for {self.__class__.__name__}.{func.__name__}: {elapsed_time}")
        return results

    return wrapper



from .download_utils import download_decorator, download_file
from .main_utils import import_module
from .lib_decorators import get_from_config, expand_input, get_elapsed_time, rgb2bgr, cast_kwargs_dict
from .decorators import get_method_time



import torch


def check_keys(model, pretrained_state_dict):
    ckpt_keys = set(pretrained_state_dict.keys())
    model_keys = set(model.state_dict().keys())
    used_pretrained_keys = model_keys & ckpt_keys
    unused_pretrained_keys = ckpt_keys - model_keys
    missing_keys = model_keys - ckpt_keys
    # print('Missing keys:{}'.format(len(missing_keys)))
    # print('Unused checkpoint keys:{}'.format(len(unused_pretrained_keys)))
    # print('Used keys:{}'.format(len(used_pretrained_keys)))
    assert len(used_pretrained_keys) > 0, 'load NONE from pretrained checkpoint'
    return True


def remove_prefix(state_dict, prefix):
    ''' Old style model is stored with all names of parameters sharing common prefix 'module.' '''
    f = lambda x: x.split(prefix, 1)[-1] if x.startswith(prefix) else x
    return {f(key): value for key, value in state_dict.items()}


def load_model(model, pretrained_path, load_to_cpu):
    if load_to_cpu:
        pretrained_dict = torch.load(pretrained_path, map_location=lambda storage, loc: storage)
    else:
        device = torch.cuda.current_device()
        pretrained_dict = torch.load(pretrained_path, map_location=lambda storage, loc: storage.cuda(device))
    if "state_dict" in pretrained_dict.keys():
        pretrained_dict = remove_prefix(pretrained_dict['state_dict'], 'module.')
    else:
        pretrained_dict = remove_prefix(pretrained_dict, 'module.')
    check_keys(model, pretrained_dict)
    model.load_state_dict(pretrained_dict, strict=False)
    return model



import tensorflow as tf
from tensorflow.keras.callbacks import Callback


class LRScalar(Callback):
    def __init__(self, round_n: int = 6):
        super().__init__()
        self.global_step = 0
        self.round_n = round_n

    def on_batch_end(self, batch, logs=None):
        try:
            lr = round(self.model.optimizer.lr.numpy(), self.round_n)
        except AttributeError:
            lr = round(float(self.model.optimizer.lr(self.global_step).numpy()), self.round_n)
        tf.summary.scalar(name='learning_rate', data=lr, step=self.global_step)
        self.global_step += 1



from deep_utils.utils.resize_utils.main_resize import resize
from .cv2_config import CV2Config
from deep_utils.utils.lib_utils.lib_decorators import get_from_config


class CV2Main:
    def __init__(self, config: CV2Config):
        self.model = None
        self.config = config

    @get_from_config
    def forward(self,
                image,
                resize_size=None,
                img_scale_factor=None,
                img_mean=None,
                swap_rgb=None,
                resize_mode=None
                ):
        import cv2
        image_len = len(image.shape)
        resize_img = resize(image, resize_size, mode=resize_mode)
        if image_len == 3:
            blobs = cv2.dnn.blobFromImage(
                resize_img,
                scalefactor=img_scale_factor,
                size=resize_size,
                mean=img_mean,
                swapRB=swap_rgb
            )
        elif image_len == 4:
            blobs = cv2.dnn.blobFromImages(
                resize_img,
                scalefactor=img_scale_factor,
                size=resize_size,
                mean=img_mean,
                swapRB=swap_rgb
            )
        else:
            raise Exception("The shape of the input image is not valid.")
        self.model.setInput(blobs)
        results = self.model.forward()
        return results



import os
import shutil
from os.path import join


def transfer_directory_items(in_dir, out_dir, transfer_list, mode='cp', remove_out_dir=False):
    print(f'starting to copying/moving from {in_dir} to {out_dir}')
    if remove_out_dir or os.path.isdir(out_dir):
        remove_create(out_dir)
    else:
        os.makedirs(out_dir, exist_ok=True)
    if mode == 'cp':
        for name in transfer_list:
            shutil.copy(os.path.join(in_dir, name), out_dir)
    elif mode == 'mv':
        for name in transfer_list:
            shutil.move(os.path.join(in_dir, name), out_dir)
    else:
        raise ValueError(f'{mode} is not supported, supported modes: mv and cp')
    print(f'finished copying/moving from {in_dir} to {out_dir}')


def dir_train_test_split(in_dir, train_dir='./train', val_dir='./val', test_size=0.1, mode='cp', remove_out_dir=False):
    from sklearn.model_selection import train_test_split
    list_ = os.listdir(in_dir)
    train_name, val_name = train_test_split(list_, test_size=test_size)
    transfer_directory_items(in_dir, train_dir, train_name, mode=mode, remove_out_dir=remove_out_dir)
    transfer_directory_items(in_dir, val_dir, val_name, mode=mode, remove_out_dir=remove_out_dir)
    return train_name, val_name


def split_dir_of_dir(in_dir, train_dir='./train', val_dir='./val', test_size=0.1, mode='cp', remove_out_dir=False):
    for data in os.listdir(in_dir):
        dir_ = join(in_dir, data)
        dir_train_test_split(dir_, train_dir=join(train_dir, data), val_dir=join(val_dir, data), mode=mode,
                             test_size=test_size, remove_out_dir=remove_out_dir)


def split_xy_dir(x_in_dir,
                 y_in_dir,
                 x_train_dir='train/samples',
                 y_train_dir='train/targets',
                 x_val_dir='val/samples',
                 y_val_dir='val/targets',
                 mode='cp',
                 val_size=0.1,
                 remove_out_dir=False):
    train_names, val_names = dir_train_test_split(x_in_dir,
                                                  train_dir=x_train_dir,
                                                  val_dir=x_val_dir,
                                                  mode=mode,
                                                  remove_out_dir=remove_out_dir,
                                                  test_size=val_size)
    train_labels = [os.path.splitext(name)[0] + '.txt' for name in train_names]
    val_labels = [os.path.splitext(name)[0] + '.txt' for name in val_names]

    transfer_directory_items(y_in_dir, y_train_dir,
                             train_labels, mode=mode, remove_out_dir=remove_out_dir)
    transfer_directory_items(y_in_dir, y_val_dir, val_labels,
                             mode=mode, remove_out_dir=remove_out_dir)


def remove_create(dir_):
    import os
    import shutil
    if os.path.exists(dir_):
        shutil.rmtree(dir_)
    os.makedirs(dir_)



import torch
from itertools import product as product
import numpy as np
from math import ceil


class PriorBox:
    def __init__(self, cfg):
        super(PriorBox, self).__init__()
        self.min_sizes = cfg['min_sizes']
        self.steps = cfg['steps']
        self.clip = cfg['clip']
        self.name = "s"

    def forward(self, image_size):
        anchors = []
        feature_maps = [[ceil(image_size[0] / step), ceil(image_size[1] / step)] for step in self.steps]
        for k, f in enumerate(feature_maps):
            min_sizes = self.min_sizes[k]
            for i, j in product(range(f[0]), range(f[1])):
                for min_size in min_sizes:
                    s_kx = min_size / image_size[1]
                    s_ky = min_size / image_size[0]
                    dense_cx = [x * self.steps[k] / image_size[1] for x in [j + 0.5]]
                    dense_cy = [y * self.steps[k] / image_size[0] for y in [i + 0.5]]
                    for cy, cx in product(dense_cy, dense_cx):
                        anchors += [cx, cy, s_kx, s_ky]
        output = torch.Tensor(anchors).view(-1, 4)
        if self.clip:
            output.clamp_(max=1, min=0)
        return output



import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models._utils as _utils

from .net import FPN as FPN
from .net import MobileNetV1 as MobileNetV1
from .net import SSH as SSH


class ClassHead(nn.Module):
    def __init__(self, inchannels=512, num_anchors=3):
        super(ClassHead, self).__init__()
        self.num_anchors = num_anchors
        self.conv1x1 = nn.Conv2d(inchannels, self.num_anchors * 2, kernel_size=(1, 1), stride=1, padding=0)

    def forward(self, x):
        out = self.conv1x1(x)
        out = out.permute(0, 2, 3, 1).contiguous()

        return out.view(out.shape[0], -1, 2)


class BboxHead(nn.Module):
    def __init__(self, inchannels=512, num_anchors=3):
        super(BboxHead, self).__init__()
        self.conv1x1 = nn.Conv2d(inchannels, num_anchors * 4, kernel_size=(1, 1), stride=1, padding=0)

    def forward(self, x):
        out = self.conv1x1(x)
        out = out.permute(0, 2, 3, 1).contiguous()

        return out.view(out.shape[0], -1, 4)


class LandmarkHead(nn.Module):
    def __init__(self, inchannels=512, num_anchors=3):
        super(LandmarkHead, self).__init__()
        self.conv1x1 = nn.Conv2d(inchannels, num_anchors * 10, kernel_size=(1, 1), stride=1, padding=0)

    def forward(self, x):
        out = self.conv1x1(x)
        out = out.permute(0, 2, 3, 1).contiguous()

        return out.view(out.shape[0], -1, 10)


class RetinaFace(nn.Module):
    def __init__(self, cfg=None, phase='test', pretrained_model=None):
        """
        :param cfg:  Network related settings.
        :param phase: train or test.
        """
        super(RetinaFace, self).__init__()
        self.phase = phase
        backbone = None
        if cfg['name'] == 'mobilenet':
            backbone = MobileNetV1()
            if cfg['pretrain']:
                checkpoint = torch.load(pretrained_model, map_location=torch.device('cpu'))
                from collections import OrderedDict
                new_state_dict = OrderedDict()
                for k, v in checkpoint['state_dict'].items():
                    name = k[7:]  # remove module.
                    new_state_dict[name] = v
                # load params
                backbone.load_state_dict(new_state_dict)
        elif cfg['name'] == 'resnet50':
            import torchvision.models as models
            backbone = models.resnet50(pretrained=cfg['pretrain'])

        self.body = _utils.IntermediateLayerGetter(backbone, cfg['return_layers'])
        in_channels_stage2 = cfg['in_channel']
        in_channels_list = [
            in_channels_stage2 * 2,
            in_channels_stage2 * 4,
            in_channels_stage2 * 8,
        ]
        out_channels = cfg['out_channel']
        self.fpn = FPN(in_channels_list, out_channels)
        self.ssh1 = SSH(out_channels, out_channels)
        self.ssh2 = SSH(out_channels, out_channels)
        self.ssh3 = SSH(out_channels, out_channels)

        self.ClassHead = self._make_class_head(fpn_num=3, inchannels=cfg['out_channel'])
        self.BboxHead = self._make_bbox_head(fpn_num=3, inchannels=cfg['out_channel'])
        self.LandmarkHead = self._make_landmark_head(fpn_num=3, inchannels=cfg['out_channel'])

    def _make_class_head(self, fpn_num=3, inchannels=64, anchor_num=2):
        classhead = nn.ModuleList()
        for i in range(fpn_num):
            classhead.append(ClassHead(inchannels, anchor_num))
        return classhead

    def _make_bbox_head(self, fpn_num=3, inchannels=64, anchor_num=2):
        bboxhead = nn.ModuleList()
        for i in range(fpn_num):
            bboxhead.append(BboxHead(inchannels, anchor_num))
        return bboxhead

    def _make_landmark_head(self, fpn_num=3, inchannels=64, anchor_num=2):
        landmarkhead = nn.ModuleList()
        for i in range(fpn_num):
            landmarkhead.append(LandmarkHead(inchannels, anchor_num))
        return landmarkhead

    def forward(self, inputs):
        out = self.body(inputs)

        # FPN
        fpn = self.fpn(out)

        # SSH
        feature1 = self.ssh1(fpn[0])
        feature2 = self.ssh2(fpn[1])
        feature3 = self.ssh3(fpn[2])
        features = [feature1, feature2, feature3]

        bbox_regressions = torch.cat([self.BboxHead[i](feature) for i, feature in enumerate(features)], dim=1)
        classifications = torch.cat([self.ClassHead[i](feature) for i, feature in enumerate(features)], dim=1)
        ldm_regressions = torch.cat([self.LandmarkHead[i](feature) for i, feature in enumerate(features)], dim=1)

        if self.phase == 'train':
            output = (bbox_regressions, classifications, ldm_regressions)
        else:
            output = (bbox_regressions, F.softmax(classifications, dim=-1), ldm_regressions)
        return output



from deep_utils.utils.lib_utils.main_utils import import_module

HaarcascadeCV2FaceDetector = import_module(
    'deep_utils.vision.face_detection.haarcascade.cv2.haarcascade_cv2_face_detection', "HaarcascadeCV2FaceDetector")



import inspect
import time
from functools import wraps
import numpy as np


def get_from_config(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        config = self.config
        arguments = inspect.getfullargspec(func)
        if arguments.defaults is not None:
            kwargs_ = {i: j for i, j in zip(arguments.args[::-1], arguments.defaults[::-1])}
        else:
            kwargs_ = dict()
        if kwargs is not None:
            kwargs_.update(kwargs)
        kwargs = kwargs_
        for key, val in kwargs.items():
            if val is None and hasattr(config, key):
                val = getattr(config, key)
                kwargs[key] = val
        return func(self, *args, **kwargs)

    return wrapper


def expand_input(dim):
    def inner_decorator(func):
        @wraps(func)
        def wrapper(self, in_, *args, **kwargs):
            if len(in_.shape) == dim + 1:
                return func(self, in_, *args, **kwargs)
            elif len(in_.shape) == dim:
                in_ = np.expand_dims(in_, axis=0)
                results = func(self, in_, *args, **kwargs)
                if type(results) is tuple:
                    results = tuple([res[0] if res is not None and len(res) == 1 else res for res in results])
                elif type(results) is dict:
                    results = {key: val[0] if val is not None and len(val) == 1 else val for key, val in
                               results.items()}
                elif hasattr(results, "DictNamedTuple") and results.DictNamedTuple:
                    new_results = {key: val[0] if val is not None and len(val) == 1 else val for key, val in
                                   results.items()}
                    cls = type(results)
                    results = cls(**new_results)
                else:
                    results = results[0]
                return results
            else:
                raise Exception(f"shape {in_.shape} is not supported.")

        return wrapper

    return inner_decorator


def get_elapsed_time(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        elapsed_time = kwargs.get('get_time', False)
        if elapsed_time:
            tic = time.time()
            results = func(self, *args, **kwargs)
            toc = time.time()
            elapsed_time = round(toc - tic, 4)
            if type(results) is dict:
                results['elapsed_time'] = elapsed_time
            elif hasattr(results, "DictNamedTuple") and results.DictNamedTuple:
                from deep_utils.utils.utils.main import dictnamedtuple
                new_results = dict(results.items())
                new_results['elapsed_time'] = elapsed_time
                cls = dictnamedtuple(results.TypeName, list(results._fields) + ["elapsed_time"])
                results = cls(**new_results)
            else:
                results = tuple(list(results) + [elapsed_time])
            return results
        return func(self, *args, **kwargs)

    return wrapper


def rgb2bgr(in_):
    def inner_decorator(func):
        @wraps(func)
        def wrapper(self, in_img, *args, **kwargs):
            is_rgb = kwargs.get('is_rgb', False)
            if not is_rgb and in_ == 'rgb':
                in_img = in_img[..., ::-1]
            elif is_rgb and in_ == 'bgr':
                in_img = in_img[..., ::-1]
            elif is_rgb and in_ == 'gray':
                in_img = np.dot(in_img[..., :3], [0.299, 0.587, 0.144])
                in_img = in_img.astype(np.uint8)
            elif not is_rgb and in_ == 'gray':
                in_img = np.dot(in_img[..., :3][..., ::-1], [0.299, 0.587, 0.144])
                in_img = in_img.astype(np.uint8)
            return func(self, in_img, *args, **kwargs)

        return wrapper

    return inner_decorator


def cast_kwargs_dict(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        arguments = inspect.getfullargspec(func)
        for kwarg in arguments.args:
            if kwarg.endswith('_kwargs'):
                _kwargs = kwargs.get(kwarg, None)
                kwargs[kwarg] = dict() if _kwargs is None else _kwargs
        return func(self, *args, **kwargs)

    return wrapper



def dump_pickle(file_path: str, file, mode: str = 'wb'):
    import pickle
    with open(file_path, mode=mode) as f:
        pickle.dump(file, f)


def load_pickle(file_path: str, mode: str = 'rb'):
    import pickle
    with open(file_path, mode=mode) as f:
        return pickle.load(f)



from deep_utils.utils.lib_utils import import_module

UltralightTorchFaceDetector = import_module(
    'deep_utils.vision.face_detection.ultralight.torch.ultralight_torch_face_detection',
    'UltralightTorchFaceDetector')
UltralightTFFaceDetector = import_module(
    'deep_utils.vision.face_detection.ultralight.tf.ultralight_tf_face_detection',
    'UltralightTFFaceDetector')




from deep_utils.utils.os_utils.os_path import split_extension


class ModelCheckPoint:
    def __init__(self,
                 model_path,
                 model,
                 monitor='min',
                 save_best_only=True,
                 overwrite=True,
                 verbose=True,
                 save_last=True,
                 loss=None,
                 optimizer=None,
                 scheduler=None,
                 static_dict=None):
        self.overwrite = overwrite
        self.model_path = model_path
        self.monitor_val = float('-inf') if monitor == 'max' else float('inf')
        self.save_best_only = save_best_only
        self.model = model
        self.loss = loss
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.static_dict = static_dict
        self.monitor = monitor
        self.epoch = 0
        self.verbose = verbose
        self.save_last = save_last

    def __call__(self, monitor_val):
        self.epoch += 1
        if self.save_best_only:
            trigger = False
            if self.monitor == 'min' and monitor_val < self.monitor_val:
                self.monitor_val = monitor_val
                trigger = True
            elif self.monitor == 'max' and monitor_val > self.monitor_val:
                self.monitor_val = monitor_val
                trigger = True

            if self.save_last:
                last_path = split_extension(self.model_path, suffix="_last")
                self._save(last_path, print_=False)

            if self.overwrite:
                best_path = split_extension(self.model_path, suffix="_best")
            else:
                best_path = split_extension(self.model_path, suffix="_" + str(self.epoch))

            if trigger:
                self._save(best_path, print_=self.verbose)
        else:
            model_path = split_extension(self.model_path, suffix="_" + str(self.epoch))
            self._save(model_path, print_=self.verbose)

    def _save(self, model_path, print_):
        import torch
        save_dict = self.static_dict if self.static_dict is not None else dict()
        save_dict['model_state_dict'] = self.model.state_dict()
        self._add_file(save_dict, 'optimizer', self.optimizer)
        self._add_file(save_dict, 'scheduler', self.scheduler)
        self._add_file(save_dict, 'loss', self.loss)
        torch.save(save_dict, model_path)
        if print_:
            print(f'model is saved in {model_path}')

    @staticmethod
    def _add_file(dict_, name, file):
        if file is not None:
            dict_[name] = file
        return dict_



import numpy as np
from deep_utils.vision.face_detection.main.main_face_detection import FaceDetector
from deep_utils.utils.lib_utils.lib_decorators import get_from_config, expand_input, get_elapsed_time, rgb2bgr
from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.box_utils.boxes import Box, Point
from .config import Config
import sys


class RetinaFaceTorchFaceDetector(FaceDetector):
    def __init__(self, **kwargs):
        import torch
        from .src.layers.functions.prior_box import PriorBox
        from .src.utils.nms.py_cpu_nms import py_cpu_nms
        from .src.retinaface import RetinaFace
        from .src.utils.box_utils import decode, decode_landm
        from .src.load_model import load_model
        self.RetinaFace = RetinaFace
        self.load_model_weights = load_model
        self.torch = torch
        self.PriorBox = PriorBox
        self.py_cpu_nms = py_cpu_nms
        self.decode = decode
        self.decode_landm = decode_landm
        self.resize = 1
        self.pretrained_mobilenet = None
        network = Config.network.lower()

        if network == 'resnet50':
            download_variables = ("resnet50",)
        elif network == "mobilenet":
            download_variables = ("mobilenet", "mobilenetV1X")
        else:
            raise ValueError('The model is not supported\nsupported models: resnet50, mobilenet')
        super().__init__(name=self.__class__.__name__,
                         file_path=__file__,
                         download_variables=download_variables,
                         **kwargs)
        self.config: Config

    @download_decorator
    def load_model(self):
        cfg_resnet = Config.cfg_re50
        cfg_mobilenet = Config.cfg_mnet
        device = self.config.device
        network = str.lower(self.config.network)
        load_to_cpu = False if device == 'cuda' else True
        if network == 'resnet50':
            cfg, model_path = (cfg_resnet, self.config.resnet50)
            net = self.RetinaFace(cfg=cfg, phase='test')
            net = self.load_model_weights(net, model_path, load_to_cpu=load_to_cpu)
        elif network == 'mobilenet':
            cfg, model_path = cfg_mobilenet, self.config.mobilenet
            net = self.RetinaFace(cfg=cfg, phase='test', pretrained_model=self.config.mobilenetV1X)
            net = self.load_model_weights(net, model_path, load_to_cpu=load_to_cpu)
        else:
            print("The network backbone is wrong!")
            sys.exit(1)
        self.config.cfg = cfg
        net.eval()
        net.to(device)
        self.config.priorbox = self.PriorBox(self.config.cfg)
        self.model = net

    @get_from_config
    @get_elapsed_time
    @rgb2bgr('bgr')
    @expand_input(3)
    def detect_faces(self,
                     images,
                     is_rgb,
                     resize_size=None,
                     img_scale_factor=None,
                     img_mean=None,
                     resize_mode=None,
                     confidence=None,
                     get_time=False,
                     round_prec=4):
        device = self.config.device
        face_points = ["left_eye", "right_eye", "nose", "mouth_left", "mouth_right"]
        images = np.float32(images)
        n_images = images.shape[0]
        images -= (104, 117, 123)
        images = images.transpose((0, 3, 1, 2))
        img = self.torch.FloatTensor(images).to(device)
        _, im_height, im_width = img[0].shape
        scale = self.torch.Tensor([img[0].shape[2], img[0].shape[1], img[0].shape[2], img[0].shape[1]])
        img = img.to(device)
        scale = scale.to(device)
        loc, conf, landms = self.model(img)

        priors = self.config.priorbox.forward(image_size=(im_height, im_width))
        priors = priors.to(device)
        prior_data = priors.data
        bboxes_, landmk_, scores_ = [], [], []
        for n_img in range(n_images):
            loc_data = loc.data[n_img]
            boxes = self.decode(loc_data, prior_data, self.config.cfg['variance'])
            boxes = boxes * scale / self.resize
            boxes = boxes.cpu().numpy()
            scores = conf[n_img].data.cpu().numpy()[:, 1]

            landms = self.decode_landm(landms.data[n_img], prior_data, self.config.cfg['variance'])
            scale1 = self.torch.Tensor([img.shape[3], img.shape[2], img.shape[3], img.shape[2],
                                        img.shape[3], img.shape[2], img.shape[3], img.shape[2],
                                        img.shape[3], img.shape[2]])
            scale1 = scale1.to(device)
            landms = landms * scale1 / self.resize
            landms = landms.cpu().numpy()

            # ignore low scores
            inds = np.where(scores > confidence)[0]
            boxes = boxes[inds]
            landms = landms[inds]
            scores = scores[inds]

            # keep top-K before NMS
            order = scores.argsort()[::-1][:self.config.top_k]
            boxes = boxes[order]
            landms = landms[order]
            scores = scores[order]

            # do NMS
            bboxes = np.hstack((boxes, scores[:, np.newaxis])).astype(np.float32, copy=False)
            keep = self.py_cpu_nms(bboxes, self.config.nms_thresholds)
            bboxes = bboxes[keep, :]
            landms = landms[keep]

            # keep top-K faster NMS
            bboxes = bboxes[:self.config.keep_top_k, :]
            landms = landms[:self.config.keep_top_k, :]
            landmk_.append(landms)
            bboxes_.append(bboxes)
            scores_.append(scores)
        bboxes, landms, scores = np.array(bboxes_), np.array(landmk_), np.array(scores_)
        boxes_, confidences_, landmarks_ = [], [], []
        for num in range(len(bboxes)):
            box_tensor = bboxes[num, :]
            boxes = [[b[0], b[1], b[2], b[3]] for b in box_tensor]
            boxes = Box.box2box(boxes, in_source=Box.BoxSource.Torch, to_source=Box.BoxSource.Numpy)
            boxes_.append(boxes)
            confidences_.append(scores.round(round_prec))
            if len(landms) != 0:
                landmarks = [[Point.point2point((landms[num][j][2 * i], landms[num][j][2 * i + 1]),
                                                in_source='Torch', to_source='Numpy') for i in range(5)] for j in
                             range(len(landms[num]))]
            img_landmarks = []
            for i in range(len(landmarks)):
                face_dict = {}
                for points, face in zip(landmarks[i], face_points):
                    face_dict[face] = points
                img_landmarks.append(face_dict)
            landmarks_.append(img_landmarks)

        output = self.output_class(boxes=boxes_, confidences=confidences_, landmarks=landmarks_[00])
        return output



from deep_utils.main_abs.cv2.cv2_config import CV2Config


class Config(CV2Config):
    def __init__(self):
        super(Config, self).__init__()
        self.prototxt_url = 'https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/deploy.prototxt.txt'
        self.prototxt_cache = 'weights/vision/face_detection/sdd/caffe/deploy.prototxt.txt'
        self.prototxt = None
        self.caffemodel_url = 'https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/res10_300x300_ssd_iter_140000.caffemodel'
        self.caffemodel_cache = 'weights/vision/face_detection/sdd/caffe/res10_300x300_ssd_iter_140000.caffemodel'
        self.caffemodel = None
        self.confidence = 0.5



import os
import shutil
import sys
from functools import wraps
import requests


def get_file(fname,
             origin,
             cache_dir=None):
    base_dir, fname = os.path.split(fname)

    if cache_dir is None:
        if 'DEEP_UTILS_HOME' in os.environ:
            cache_dir = os.environ.get('DEEP_UTILS_HOME')
        else:
            cache_dir = os.path.join(os.path.expanduser('~'), '.deep_utils')

    datadir_base = os.path.expanduser(cache_dir)
    datadir = os.path.join(datadir_base, base_dir)
    os.makedirs(datadir, exist_ok=True)

    fpath = os.path.join(datadir, fname)
    if fpath.endswith(".zip"):
        check_path = fpath[:-4]
    else:
        check_path = fpath
    download = False if os.path.exists(check_path) else True

    if download:
        print('Downloading data from', origin)
        error_msg = 'URL fetch failure on {}'
        try:
            with open(fpath, 'wb') as f:
                response = requests.get(origin, stream=True)
                total = response.headers.get('content-length')

                if total is None:
                    f.write(response.content)
                else:
                    downloaded = 0
                    total = int(total)
                    for data in response.iter_content(chunk_size=max(int(total / 1000), 1024 * 1024)):
                        downloaded += len(data)
                        f.write(data)
                        done = int(50 * downloaded / total)
                        sys.stdout.write('\r[{}{}]'.format('█' * done, '.' * (50 - done)))
                        sys.stdout.flush()
            sys.stdout.write('\n')
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise Exception(error_msg.format(origin))
        if fpath.endswith(".zip"):
            from zipfile import ZipFile
            try:
                with ZipFile(fpath, 'r') as zip:
                    zip.printdir()
                    print(f'extracting {fpath}')
                    zip.extractall(fpath[:-4])
                    print(f'extracting is done!')
                os.remove(fpath)
            except (Exception, KeyboardInterrupt):
                raise Exception(f"Error in extracting {fpath}")
    if fpath.endswith('.zip'):
        fpath = fpath[:-4]
    return fpath


def download_file(url, download_dir='.', unzip=False, remove_zip=False, file_name=None):
    if url is None:
        print('url is None. Exiting the function')
        return
    os.makedirs(download_dir, exist_ok=True)
    print('Downloading data from', url)
    error_msg = 'URL fetch failure on {}'

    try:
        response = requests.get(url, stream=True)
        total = response.headers.get('content-length')
        while not total:
            response = requests.get(url, stream=True)
            total = response.headers.get('content-length')
        try:
            if file_name is None:
                file_name = response.headers.get('filename')
                if file_name is None:
                    file_name = response.headers.get('content-disposition').split('=')[-1]
        except:
            file_name = os.path.split(url)[-1]

        if not os.path.isdir(download_dir):
            os.makedirs(download_dir, exist_ok=True)
        download_des = os.path.join(download_dir, file_name)
        with open(download_des, 'wb') as f:
            if total is None:
                f.write(response.content)
            else:
                downloaded = 0
                total = int(total)
                for data in response.iter_content(chunk_size=max(int(total / 1000), 1024 * 1024)):
                    downloaded += len(data)
                    f.write(data)
                    done = int(50 * downloaded / total)
                    sys.stdout.write('\r[{}{}]'.format('█' * done, '.' * (50 - done)))
                    sys.stdout.flush()
        sys.stdout.write('\n')
    except (Exception, KeyboardInterrupt):
        if os.path.isfile(download_des):
            os.remove(download_des)
        raise Exception(error_msg.format(url))
    if unzip:
        from zipfile import ZipFile
        with ZipFile(download_des, 'r') as zip:
            print(f'extracting {download_des}')
            zip.extractall(download_dir)
            print(f'extracting is done!')
        if remove_zip:
            os.remove(download_des)


def download_decorator(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        config = self.config
        download_variables = self.download_variables
        for var in download_variables:
            if getattr(config, var) is None:
                url = getattr(config, var + '_url')
                f_name = getattr(config, var + '_cache')
                f_path = get_file(f_name, url)
                setattr(config, var, f_path)
        return func(self, *args, **kwargs)

    return wrapper


if __name__ == '__main__':
    url = 'https://isic-challenge-data.s3.amazonaws.com/2016/ISBI2016_ISIC_Part1_Test_GroundTruth.zip'
    download_file(url, '/home/ai/pooya', unzip=True, remove_zip=False)
    download_file(url, '/home/ai/pooya', unzip=True, remove_zip=False, file_name='wow.zip')
    download_file("https://github.com/Practical-AI/deep_utils/archive/refs/tags/0.4.2.zip", '/home/ai/pooya',
                  unzip=True, remove_zip=True)
    download_file("https://github.com/Practical-AI/deep_utils/archive/refs/tags/0.4.2.zip", '/home/ai/pooya',
                  unzip=False, remove_zip=True)



import inspect
import os
from abc import ABC
from typing import Union
from deep_utils.utils.lib_utils.main_utils import import_module
from deep_utils.utils.os_utils import split_all


class MainClass(ABC):
    def __init__(self, name, file_path, **kwargs):
        self.config = None
        self.model = None
        self.name = name
        self.download_variables: Union[tuple, None] = kwargs.get('download_variables', None)
        self.load_config(file_path, **kwargs)
        self.load_model()

    def __repr__(self):
        message = f"{self.name} config attributes:"
        for name, val in inspect.getmembers(self.config):

            if not name.startswith('_'):
                if not inspect.ismethod(val):
                    message += f"\n{name} = {val}"
        return message

    def load_model(self):
        raise NotImplementedError('load_model is not implemented')

    def load_config(self, file_path, **kwargs):
        separated_files = os.path.join('deep_utils', os.path.split(file_path)[0].split('deep_utils')[-1][1:])
        file_ = '.'.join(split_all(separated_files) + ["config"])
        config = import_module(file_, 'Config')
        self.config = config()

        for arg, val in kwargs.items():
            try:
                if type(arg) is dict:
                    getattr(self.config, arg).update(val)
                else:
                    setattr(self.config, arg, val)
            except:
                pass



from deep_utils.utils.lib_utils import import_module

SSDCV2CaffeFaceDetector = import_module('deep_utils.vision.face_detection.ssd.cv2.caffe.ssd_cv2_caffe_face_detection',
                                        'SSDCV2CaffeFaceDetector')



from typing import Union

import numpy as np
from enum import Enum


class Point:
    class PointSource(Enum):
        Torch = 'Torch'
        TF = "TF"
        CV = 'CV'
        Numpy = 'Numpy'

    @staticmethod
    def point2point(point, in_source, to_source, in_relative=None, to_relative=None, shape=None, shape_source=None):
        if point is None or len(point) == 0:
            pass
        elif type(point[0]) in [tuple, list, np.ndarray]:
            point = [Point._point2point(p, in_source=in_source, to_source=to_source,
                                        in_relative=in_relative, to_relative=to_relative,
                                        shape=shape, shape_source=shape_source) for p in point]
        else:
            point = Point._point2point(point, in_source=in_source, to_source=to_source,
                                       in_relative=in_relative, to_relative=to_relative,
                                       shape=shape, shape_source=shape_source)
        return point

    @staticmethod
    def _point2point(point, in_source, to_source, in_relative=None, to_relative=None, shape=None, shape_source=None):
        if type(in_source) is Point.PointSource:
            in_source = in_source.value
        if type(to_source) is Point.PointSource:
            to_source = to_source.value

        if (in_source in [Point.PointSource.Torch.value, Point.PointSource.CV.value] and to_source in [
            Point.PointSource.TF.value, Point.PointSource.Numpy.value]) \
                or (in_source in [Point.PointSource.TF.value, Point.PointSource.Numpy.value] and to_source in [
            Point.PointSource.Torch.value, Point.PointSource.CV.value]):
            point = (point[1], point[0])
        elif (in_source is None and to_source is None) or in_source == to_source \
                or (in_source in [Point.PointSource.Torch.value, Point.PointSource.CV.value] and to_source in [
            Point.PointSource.CV.value, Point.PointSource.Torch.value]) \
                or (in_source in [Point.PointSource.TF.value, Point.PointSource.Numpy.value] and to_source in [
            Point.PointSource.TF.value, Point.PointSource.Numpy.value]):
            pass
        else:
            raise Exception(
                f'Conversion form {in_source} to {to_source} is not Supported.'
                f' Supported types: {Box._get_enum_names(Point.PointSource)}')
        if to_source is not None and shape_source is not None and shape is not None:
            img_w, img_h = Point.point2point(shape, in_source=shape_source, to_source=to_source)
            if not in_relative and to_relative:
                p1, p2 = point
                point = [p1 / img_w, p2 / img_h]
            elif in_relative and not to_relative:
                p1, p2 = point
                point = [p1 * img_w, p2 * img_h]
        return point

    @staticmethod
    def _put_point(img, point, radius, color=(0, 255, 0), thickness=None, lineType=None, shift=None, in_source="Numpy"):
        import cv2
        if type(point) is not int:
            point = (int(point[0]), int(point[1]))
        point = Point.point2point(point, in_source=in_source, to_source="CV")
        return cv2.circle(img, point, radius, color, thickness, lineType, shift)

    @staticmethod
    def put_point(img, point, radius, color=(0, 255, 0), thickness=None, lineType=None, shift=None, in_source="Numpy"):
        if point is None or len(point) == 0:
            pass
        elif type(point[0]) in [tuple, list, np.ndarray]:
            for p in point:
                img = Point._put_point(img, p, radius, color, thickness, lineType, shift, in_source)
        else:
            img = Point._put_point(img, point, radius, color, thickness, lineType, shift, in_source)
        return img


class Box:
    class BoxFormat(Enum):
        XYWH = "XYWH"
        XYXY = "XYXY"
        XCYC = "XCYC"

    class BoxSource(Enum):
        Torch = 'Torch'
        TF = "TF"
        CV = 'CV'
        Numpy = 'Numpy'

    class OutType(Enum):
        Numpy = np.array
        List = list
        Tuple = tuple

    @staticmethod
    def box2box(box,
                in_format=None,
                to_format=None,
                in_source=BoxSource.Numpy,
                to_source=BoxSource.Numpy,
                in_relative=None,
                to_relative=None,
                shape=None,
                shape_source=None,
                out_type=None,
                return_int=None):
        if box is None or len(box) == 0:
            pass
        elif type(box[0]) in [tuple, list, np.ndarray]:
            box = [Box._box2box(b, in_format=in_format, to_format=to_format, in_source=in_source, to_source=to_source,
                                in_relative=in_relative, to_relative=to_relative, shape=shape,
                                shape_source=shape_source, out_type=out_type, return_int=return_int)
                   for b in box]

        else:
            box = Box._box2box(box, in_format=in_format, to_format=to_format, in_source=in_source, to_source=to_source,
                               in_relative=in_relative, to_relative=to_relative, shape=shape,
                               shape_source=shape_source,
                               out_type=out_type, return_int=return_int)
        return box

    @staticmethod
    def _box2box(box,
                 in_format=None,
                 to_format=None,
                 in_source=None,
                 to_source=None,
                 in_relative=None,
                 to_relative=None,
                 shape=None,
                 shape_source=None,
                 out_type=None,
                 return_int=None):
        """

        :param box:
        :param in_format:
        :param to_format:
        :param in_source:
        :param to_source:
        :param relative:
        :param img_w:
        :param img_h:
        :param out_type: output type of the box. Supported types: list, tuple, numpy
        :return:
        """
        if type(in_format) is Box.BoxFormat:
            in_format = in_format.value
        if type(to_format) is Box.BoxFormat:
            to_format = to_format.value

        if type(in_source) is Box.BoxSource:
            in_source = in_source.value
        if type(to_source) is Box.BoxSource:
            to_source = to_source.value

        if in_format == Box.BoxFormat.XYWH.value and to_format == Box.BoxFormat.XYXY.value:
            x1, y1, w, h = box
            x2, y2 = x1 + w, y1 + h
            box = [x1, y1, x2, y2]
        elif in_format == Box.BoxFormat.XYXY.value and to_format == Box.BoxFormat.XYWH.value:
            x1, y1, x2, y2 = box
            w, h = x2 - x1, y2 - y1
            box = [x1, y1, w, h]
        elif in_format == Box.BoxFormat.XYXY.value and to_format == Box.BoxFormat.XCYC.value:
            x1, y1, x2, y2 = box
            w, h = x2 - x1, y2 - y1
            xc, yc = (x1 + x2) / 2, (y1 + y2) / 2
            box = [xc, yc, w, h]
        elif in_format == Box.BoxFormat.XCYC.value and to_format == Box.BoxFormat.XYXY.value:
            xc, yc, w, h = box
            x1, y1, x2, y2 = xc - w / 2, yc - h / 2, xc + w / 2, yc + h / 2
            box = [x1, y1, x2, y2]
        elif in_format == Box.BoxFormat.XYWH.value and to_format == Box.BoxFormat.XCYC.value:
            x1, y1, w, h = box
            x2, y2 = x1 + w, y1 + h
            xc, yc = (x1 + x2) / 2, (y1 + y2) / 2
            box = [xc, yc, w, h]
        elif in_format == Box.BoxFormat.XCYC.value and to_format == Box.BoxFormat.XYWH.value:
            xc, yc, w, h = box
            x1, y1 = xc - w // 2, yc - h // 2

            box = [x1, y1, w, h]
        elif (in_format is None and to_format is None) or in_format == to_format:
            pass
        else:
            raise Exception(
                f'Conversion form {in_format} to {to_format} is not Supported.'
                f' Supported types: {Box._get_enum_names(Box.BoxFormat)}')

        if (in_source in [Box.BoxSource.Torch.value, Box.BoxSource.CV.value] and to_source in [
            Box.BoxSource.TF.value, Box.BoxSource.Numpy.value]) \
                or (in_source in [Box.BoxSource.TF.value, Box.BoxSource.Numpy.value] and to_source in [
            Box.BoxSource.Torch.value, Box.BoxSource.CV.value]):
            box = [box[1], box[0], box[3], box[2]]
        elif (in_source is None and to_source is None) or in_source == to_source \
                or (in_source in [Box.BoxSource.Torch.value, Box.BoxSource.CV.value] and to_source in [
            Box.BoxSource.CV.value, Box.BoxSource.Torch.value]) \
                or (in_source in [Box.BoxSource.TF.value, Box.BoxSource.Numpy.value] and to_source in [
            Box.BoxSource.TF.value, Box.BoxSource.Numpy.value]):
            pass
        else:
            raise Exception(
                f'Conversion form {in_source} to {to_source} is not Supported.'
                f' Supported types: {Box._get_enum_names(Box.BoxSource)}')
        if to_source is not None and shape_source is not None and shape is not None:
            img_w, img_h = Point.point2point(shape, in_source=shape_source, to_source=to_source)
            if not in_relative and to_relative:
                b1, b2, b3, b4 = box
                box = [b1 / img_w, b2 / img_h, b3 / img_w, b4 / img_h]
            elif in_relative and not to_relative:
                b1, b2, b3, b4 = box
                box = [b1 * img_w, b2 * img_h, b3 * img_w, b4 * img_h]

        box = Box.get_type(box, out_type)
        if return_int:
            box = [int(b) for b in box]
        return box

    @staticmethod
    def get_type(in_, out_type):
        if out_type is not None:
            try:
                in_ = out_type(in_)
            except:
                raise Exception(
                    f'{out_type} is not Supported. Supported types: {Box._get_enum_names(Box.OutType)}')
        return in_

    @staticmethod
    def _get_enum_names(in_):
        return [n.name for n in in_]

    @staticmethod
    def _put_box(img, box, copy=False,
                 color=(0, 255, 0),
                 thickness=1,
                 lineType=None,
                 shift=None,
                 in_relative=False,
                 in_format="XYXY",
                 in_source='Numpy'):
        import cv2
        box = Box.box2box(box,
                          in_format=in_format,
                          to_format=Box.BoxFormat.XYXY,
                          in_source=in_source,
                          to_source=Box.BoxSource.CV,
                          in_relative=in_relative,
                          to_relative=False,
                          shape=img.shape[:2],
                          shape_source='Numpy')

        if type(img) is not np.ndarray:
            img = np.array(img).astype(np.uint8)
        else:
            img = img.astype(np.uint8)
        box = [int(point) for point in box]
        if copy:
            img = img.copy()
        img = cv2.rectangle(img, (box[0], box[1]), (box[2], box[3]), color=color, thickness=thickness,
                            lineType=lineType,
                            shift=shift)
        return img

    @staticmethod
    def put_box(img, box, copy=False,
                color=(0, 255, 0),
                thickness=1,
                lineType=None,
                shift=None,
                in_relative=False,
                in_format=BoxFormat.XYXY,
                in_source=BoxSource.Numpy):
        if box is None or len(box) == 0:
            pass
        elif type(box[0]) in [tuple, list, np.ndarray]:
            for b in box:
                img = Box._put_box(img, box=b, copy=copy, color=color, thickness=thickness, lineType=lineType,
                                   shift=shift, in_format=in_format, in_source=in_source, in_relative=in_relative)
        else:
            img = Box._put_box(img, box=box, copy=copy, color=color, thickness=thickness, lineType=lineType,
                               shift=shift, in_format=in_format, in_source=in_source, in_relative=in_relative)

        return img

    @staticmethod
    def _get_box_img(img, bbox, box_format=BoxFormat.XYXY, box_source=BoxSource.Numpy):
        bbox = Box.box2box(bbox, in_format=box_format, to_format=Box.BoxFormat.XYXY, in_source=box_source,
                           to_source=Box.BoxSource.Numpy, return_int=True)
        img_part = img[bbox[0]:bbox[2], bbox[1]:bbox[3]]
        return img_part

    @staticmethod
    def get_box_img(img, bbox, box_format=BoxFormat.XYXY, box_source=BoxSource.Numpy):
        if len(img.shape) != 3:
            raise Exception('The image size should be 3')

        img_part = []
        if bbox is None or len(bbox) == 0:
            pass
        elif type(bbox[0]) in [tuple, list, np.ndarray]:
            img_part = [Box._get_box_img(img, b, box_format, box_source) for b in bbox]
        else:
            img_part = Box._get_box_img(img, bbox, box_format, box_source)
        return img_part

    @staticmethod
    def _put_text(img, text, org, fontFace=None, fontScale=1, color=(0, 255, 0),
                  thickness=1, lineType=None, bottomLeftOrigin=None, org_source='Numpy'):
        import cv2
        org = (int(org[0]), int(org[1]))
        org = Point.point2point(org, in_source=org_source, to_source=Point.PointSource.CV)
        font_face = cv2.FONT_HERSHEY_PLAIN if fontFace is None else fontFace
        img = cv2.putText(img, text, org, font_face, fontScale, color, thickness, lineType, bottomLeftOrigin)

        return img

    @staticmethod
    def put_text(img,
                 text,
                 org,
                 fontFace=None,
                 fontScale=1,
                 color=(0, 255, 0),
                 thickness=1,
                 lineType=None,
                 bottomLeftOrigin=None,
                 org_source='Numpy'):
        if text is None or len(text) == 0 or org is None or len(org) == 0:
            pass
        elif type(text) in [tuple, list, np.ndarray]:
            for t, o in zip(text, org):
                img = Box._put_text(img, t, o, fontFace, fontScale, color, thickness, lineType, bottomLeftOrigin,
                                    org_source=org_source)
        else:
            img = Box._put_text(img, text, org, fontFace, fontScale, color, thickness, lineType, bottomLeftOrigin,
                                org_source=org_source)
        return img

    @staticmethod
    def get_biggest(box,
                    in_format=BoxFormat.XYXY,
                    in_source=BoxSource.Numpy,
                    get_index=False,
                    inputs: Union[None, dict] = None,
                    reverse=False):
        if len(box) == 0 or box is None:
            return
        box = Box.box2box(box,
                          in_format=in_format,
                          in_source=in_source,
                          to_source=Box.BoxSource.Numpy,
                          to_format=Box.BoxFormat.XYWH
                          )
        if reverse:
            chosen_box = min(box, key=lambda b: b[2] * b[3])
        else:
            chosen_box = max(box, key=lambda b: b[2] * b[3])
        index = box.index(chosen_box)
        if inputs is not None:
            inputs = {k: v[index] for k, v in inputs.items()}
            return inputs
        chosen_box = Box.box2box(chosen_box, in_format=Box.BoxFormat.XYWH, to_format=Box.BoxFormat.XYXY)
        if get_index:
            return chosen_box, index
        return chosen_box

    @staticmethod
    def get_area(box,
                 in_format=BoxFormat.XYXY,
                 in_source=BoxSource.Numpy):
        box = Box.box2box(box,
                          in_format=in_format,
                          in_source=in_source,
                          to_source=Box.BoxSource.Numpy,
                          to_format=Box.BoxFormat.XYWH
                          )
        area = box[2] * box[3]
        return area


if __name__ == '__main__':
    print(Box.BoxFormat.XYXY is Box.BoxFormat)



import os
from collections import OrderedDict
import numpy as np


def get_nets():
    import torch
    import torch.nn as nn
    import torch.nn.functional as F
    class Flatten(nn.Module):

        def __init__(self):
            super(Flatten, self).__init__()

        def forward(self, x):
            """
            Arguments:
                x: a float tensor with shape [batch_size, c, h, w].
            Returns:
                a float tensor with shape [batch_size, c*h*w].
            """

            # without this pretrained model isn't working
            x = x.transpose(3, 2).contiguous()

            return x.view(x.size(0), -1)

    class PNet(nn.Module):

        def __init__(self, pnet_path):
            super(PNet, self).__init__()

            # suppose we have input with size HxW, then
            # after first layer: H - 2,
            # after pool: ceil((H - 2)/2),
            # after second conv: ceil((H - 2)/2) - 2,
            # after last conv: ceil((H - 2)/2) - 4,
            # and the same for W

            self.features = nn.Sequential(OrderedDict([
                ('conv1', nn.Conv2d(3, 10, 3, 1)),
                ('prelu1', nn.PReLU(10)),
                ('pool1', nn.MaxPool2d(2, 2, ceil_mode=True)),

                ('conv2', nn.Conv2d(10, 16, 3, 1)),
                ('prelu2', nn.PReLU(16)),

                ('conv3', nn.Conv2d(16, 32, 3, 1)),
                ('prelu3', nn.PReLU(32))
            ]))

            self.conv4_1 = nn.Conv2d(32, 2, 1, 1)
            self.conv4_2 = nn.Conv2d(32, 4, 1, 1)

            weights = np.load(os.path.join(os.path.split(__file__)[0], pnet_path), allow_pickle=True)[()]
            for n, p in self.named_parameters():
                p.data = torch.FloatTensor(weights[n])

        def forward(self, x):
            """
            Arguments:
                x: a float tensor with shape [batch_size, 3, h, w].
            Returns:
                b: a float tensor with shape [batch_size, 4, h', w'].
                a: a float tensor with shape [batch_size, 2, h', w'].
            """
            x = self.features(x)
            a = self.conv4_1(x)
            b = self.conv4_2(x)
            a = F.softmax(a)
            return b, a

    class RNet(nn.Module):

        def __init__(self, rnet_path):
            super(RNet, self).__init__()

            self.features = nn.Sequential(OrderedDict([
                ('conv1', nn.Conv2d(3, 28, 3, 1)),
                ('prelu1', nn.PReLU(28)),
                ('pool1', nn.MaxPool2d(3, 2, ceil_mode=True)),

                ('conv2', nn.Conv2d(28, 48, 3, 1)),
                ('prelu2', nn.PReLU(48)),
                ('pool2', nn.MaxPool2d(3, 2, ceil_mode=True)),

                ('conv3', nn.Conv2d(48, 64, 2, 1)),
                ('prelu3', nn.PReLU(64)),

                ('flatten', Flatten()),
                ('conv4', nn.Linear(576, 128)),
                ('prelu4', nn.PReLU(128))
            ]))

            self.conv5_1 = nn.Linear(128, 2)
            self.conv5_2 = nn.Linear(128, 4)

            weights = np.load(os.path.join(os.path.split(__file__)[0], rnet_path), allow_pickle=True)[()]
            for n, p in self.named_parameters():
                p.data = torch.FloatTensor(weights[n])

        def forward(self, x):
            """
            Arguments:
                x: a float tensor with shape [batch_size, 3, h, w].
            Returns:
                b: a float tensor with shape [batch_size, 4].
                a: a float tensor with shape [batch_size, 2].
            """
            x = self.features(x)
            a = self.conv5_1(x)
            b = self.conv5_2(x)
            a = F.softmax(a)
            return b, a

    class ONet(nn.Module):

        def __init__(self, onet_path):
            super(ONet, self).__init__()

            self.features = nn.Sequential(OrderedDict([
                ('conv1', nn.Conv2d(3, 32, 3, 1)),
                ('prelu1', nn.PReLU(32)),
                ('pool1', nn.MaxPool2d(3, 2, ceil_mode=True)),

                ('conv2', nn.Conv2d(32, 64, 3, 1)),
                ('prelu2', nn.PReLU(64)),
                ('pool2', nn.MaxPool2d(3, 2, ceil_mode=True)),

                ('conv3', nn.Conv2d(64, 64, 3, 1)),
                ('prelu3', nn.PReLU(64)),
                ('pool3', nn.MaxPool2d(2, 2, ceil_mode=True)),

                ('conv4', nn.Conv2d(64, 128, 2, 1)),
                ('prelu4', nn.PReLU(128)),

                ('flatten', Flatten()),
                ('conv5', nn.Linear(1152, 256)),
                ('drop5', nn.Dropout(0.25)),
                ('prelu5', nn.PReLU(256)),
            ]))

            self.conv6_1 = nn.Linear(256, 2)
            self.conv6_2 = nn.Linear(256, 4)
            self.conv6_3 = nn.Linear(256, 10)

            weights = np.load(os.path.join(os.path.split(__file__)[0], onet_path), allow_pickle=True)[()]
            for n, p in self.named_parameters():
                p.data = torch.FloatTensor(weights[n])

        def forward(self, x):
            """
            Arguments:
                x: a float tensor with shape [batch_size, 3, h, w].
            Returns:
                c: a float tensor with shape [batch_size, 10].
                b: a float tensor with shape [batch_size, 4].
                a: a float tensor with shape [batch_size, 2].
            """
            x = self.features(x)
            a = self.conv6_1(x)
            b = self.conv6_2(x)
            c = self.conv6_3(x)
            a = F.softmax(a)
            return c, b, a

    return PNet, RNet, ONet



from .yolo import *
from .main import *
from deep_utils.utils.lib_utils.main_utils import list_utils, loader

Object_Detection_Models = {
    "YOLOV5TorchObjectDetector": YOLOV5TorchObjectDetector
}

list_object_detection_models = list_utils(Object_Detection_Models)


def object_detector_loader(name, **kwargs) -> ObjectDetector:
    return loader(Object_Detection_Models, list_object_detection_models)(name, **kwargs)



class Config:
    def __init__(self):
        self.haarcascade_frontalface_url = "https://raw.githubusercontent.com/opencv/opencv/master/data/haarcascades/haarcascade_frontalface_alt2.xml"
        self.haarcascade_frontalface_cache = "weights/vision/face_detection/haarcascade/cv2/haarcascade_frontalface_alt2.xml"
        self.haarcascade_frontalface = None
        self.haarcascade_eye_url = "https://raw.githubusercontent.com/opencv/opencv/master/data/haarcascades/haarcascade_eye.xml"
        self.haarcascade_eye_cache = "weights/vision/face_detection/haarcascade/cv2/haarcascade_eye.xml"
        self.haarcascade_eye = None
        self.haarcascade_nose_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/haarcascade_nose.xml"
        self.haarcascade_nose_cache = "weights/vision/face_detection/haarcascade/cv2/haarcascade_nose.xml"
        self.haarcascade_nose = None
        self.landmarks_url = "https://github.com/kurnianggoro/GSOC2017/raw/master/data/lbfmodel.yaml"
        self.landmarks_cache = "weights/vision/face_detection/haarcascade/cv2/lbfmodel.yaml"
        self.landmarks = None

        self.scale = None
        self.scaleFactor = None
        self.minSize = None
        self.minNeighbors = None
        self.flags = None
        self.maxSize = None



import numpy as np
from deep_utils.utils.box_utils.boxes import Point


def resize(img, dsize, in_source='Numpy', mode='cv2', interpolation=None):
    mode = 'cv2' if mode is None else mode
    if mode == 'cv2':
        dsize = Point.point2point(dsize, in_source=in_source, to_source=Point.PointSource.CV)
        new_img = cv2_resize(img, dsize=dsize, interpolation=interpolation)
    elif mode.lower() == 'pil':
        from PIL import Image
        img = img if type(img) == Image.Image else Image.fromarray(img)
        new_img = img.resize(dsize, interpolation)
    else:
        raise ValueError(f'{mode} is not supported, supported types: cv2, ')
    return new_img


def cv2_resize(img, dsize, dst=None, fx=None, fy=None, interpolation=None):
    import cv2
    if len(img.shape) == 3:
        return cv2.resize(img, dsize, dst, fx, fy, interpolation)
    elif len(img.shape) == 4:
        return np.array([cv2.resize(im, dsize, dst, fx, fy, interpolation) for im in img])


def get_img_shape(img):
    if len(img.shape) == 4:
        return img.shape
    elif len(img.shape) == 3:
        return np.expand_dims(img, axis=0).shape
    else:
        raise Exception(f'shape: {img.shape} is not an image')


def resize_save_aspect_ratio(img, size):
    from PIL import Image
    from torchvision import transforms
    import torch.nn.functional as F

    w, h = img.size
    ratio = h / w
    if ratio > 1:
        h = size
        w = int(size / ratio)
    elif ratio < 1:
        h = int(size * ratio)
        w = size
    else:
        return img.resize((size, size), Image.NEAREST)

    image = img.resize((w, h), Image.NEAREST)
    im = transforms.ToTensor()(image)
    val = h - w
    if val < 0:
        a = 0
        b = int(-val / 2)
    else:
        a = int(val / 2)
        b = 0

    result = F.pad(input=im, pad=(a, a, b, b))
    image = transforms.ToPILImage()(result)

    return image



import os
from os.path import join
import sys
import numpy as np
from deep_utils.utils.lib_utils.lib_decorators import get_from_config, expand_input, get_elapsed_time, rgb2bgr
from deep_utils.vision.object_detection.main.main_object_detection import ObjectDetector
from deep_utils.utils.box_utils.boxes import Box, Point
from deep_utils.utils.os_utils.os_path import split_extension
from deep_utils.utils.dir_utils.main import dir_train_test_split, transfer_directory_items
from deep_utils.utils.dir_utils.main import remove_create
from deep_utils.utils.opencv.main import show_destroy_cv2
from .config import Config


class YOLOV5TorchObjectDetector(ObjectDetector):
    def __init__(self, **kwargs):
        super(YOLOV5TorchObjectDetector, self).__init__(name=self.__class__.__name__, file_path=__file__, **kwargs)
        self.config: Config

    def load_model(self):
        import torch
        sys.path.append(os.path.split(__file__)[0])
        from .models.experimental import attempt_load
        if self.config.model_weight:
            self.model = attempt_load(self.config.model_weight, map_location=self.config.device)
            self.model.to(self.config.device)
            self.model.eval()
            img = torch.zeros((1, 3, *self.config.img_size), device=self.config.device)
            self.model(img)
            print(f'{self.name}: weights are loaded')

    @rgb2bgr('rgb')
    @get_elapsed_time
    @expand_input(3)
    @get_from_config
    def detect_objects(self,
                       img,
                       is_rgb,
                       confidence=None,
                       iou_thresh=None,
                       classes=None,
                       agnostic=None,
                       get_time=False,
                       img_size=None,
                       **kwargs
                       ):
        import torch
        from .utils.datasets import letterbox
        from .utils.general import non_max_suppression, scale_coords
        im0 = img
        img = np.array([letterbox(im, new_shape=img_size)[0] for im in im0])
        img = img.transpose((0, 3, 1, 2))
        img = np.ascontiguousarray(img)
        img = torch.from_numpy(img).to(self.config.device)
        img = img / 255.0
        with torch.no_grad():
            prediction = self.model(img, augment=False)[0]
        prediction = non_max_suppression(prediction, confidence, iou_thresh, classes=classes, agnostic=agnostic)
        self.boxes, self.class_names, self.classes, self.confidences = [[[] for _ in range(im0.shape[0])] for _ in
                                                                        range(4)]
        for i, det in enumerate(prediction):  # detections per image
            if len(det):
                det[:, :4] = scale_coords(img.shape[2:], det[:, :4], im0[i].shape).round()
                for *xyxy, conf, cls in reversed(det):
                    bbox = Box.box2box(xyxy,
                                       in_source=Box.BoxSource.Torch,
                                       to_source=Box.BoxSource.Numpy,
                                       return_int=True)
                    self.boxes[i].append(bbox)
                    self.confidences[i].append(round(conf.item(), 2))
                    cls = int(cls.item())
                    self.classes[i].append(cls)
                    self.class_names[i].append(self.config.class_names[cls])
        return dict(
            classes=self.classes,
            boxes=self.boxes,
            confidences=self.confidences,
            class_names=self.class_names
        )

    @staticmethod
    def test_label(img_path, str_path):
        import cv2
        img = cv2.imread(img_path)
        boxes, texts, orgs = [], [], []
        with open(str_path, mode='r') as f:
            for line in f.readlines():
                label, xc, yc, w, h = line.strip().split()
                xc, yc, w, h = float(xc), float(yc), float(w), float(h)
                boxes.append([xc, yc, w, h])
                texts.append(f"label: {label}")
                org = Point.point2point((xc - w / 2, yc - h / 2),
                                        in_source='CV',
                                        to_source='Numpy',
                                        in_relative=True,
                                        to_relative=False,
                                        shape_source='Numpy',
                                        shape=img.shape[:2])
                orgs.append(org)
        img = Box.put_box(img, boxes, in_source='CV', in_format=Box.BoxFormat.XCYC, in_relative=True)
        img = Box.put_text(img, texts, org=orgs, thickness=3, fontScale=3)
        show_destroy_cv2(img)
        return img

    @staticmethod
    def split_dataset(base_dir, out_dir, mode='cp', remove_out_dir=False):
        img_train_names, img_val_names = dir_train_test_split(join(base_dir, 'images'),
                                                              train_dir=join(out_dir, 'train', 'images'),
                                                              val_dir=join(out_dir, 'val', 'images'),
                                                              mode=mode,
                                                              remove_out_dir=remove_out_dir)
        img_train_labels = [os.path.splitext(name)[0] + '.txt' for name in img_train_names]
        img_val_labels = [os.path.splitext(name)[0] + '.txt' for name in img_val_names]

        transfer_directory_items(join(base_dir, 'labels'), join(out_dir, 'train', 'labels'),
                                 img_train_labels, mode=mode, remove_out_dir=remove_out_dir)
        transfer_directory_items(join(base_dir, 'labels'), join(out_dir, 'val', 'labels'), img_val_labels,
                                 mode=mode, remove_out_dir=remove_out_dir)

    @staticmethod
    def extract_label(label_path, img_path=None, shape=None, shape_source=None):
        with open(label_path, mode='r') as f:
            boxes, labels = [], []
            for line in f.readlines():
                label, b1, b2, b3, b4 = line.strip().split()
                boxes.append([float(b1), float(b2), float(b3), float(b4)])
                labels.append(int(label))

        if img_path is not None:
            import cv2
            shape = cv2.imread(img_path).shape[:2]
            shape_source = 'Numpy'
        if shape is not None and shape_source is not None:
            boxes = Box.box2box(boxes,
                                in_source=Box.BoxSource.CV,
                                to_source='Numpy',
                                in_format='XCYC',
                                to_format='XYXY',
                                in_relative=True,
                                to_relative=False,
                                shape=shape,
                                shape_source=shape_source)
        return boxes, labels

    @get_from_config
    def detect_dir(self,
                   dir_,
                   confidence=None,
                   iou_thresh=None,
                   classes=None,
                   agnostic=None,
                   img_size=None,
                   extensions=('.png', '.jpg', '.jpeg'),
                   res_img_dir=None,
                   res_label_dir=None,
                   put_annotations=True,
                   remove_dirs=False,
                   **kwargs):
        import cv2
        results = dict()
        if res_label_dir and remove_dirs:
            remove_create(res_label_dir)
        if res_img_dir and remove_dirs:
            remove_create(res_img_dir)
        for item_name in os.listdir(dir_):
            _, extension = os.path.splitext(item_name)
            if extension in extensions:
                img_path = os.path.join(dir_, item_name)
                img = cv2.imread(img_path)
                result = self.detect_objects(img,
                                             is_rgb=False,
                                             confidence=confidence,
                                             iou_thresh=iou_thresh,
                                             classes=classes,
                                             agnostic=agnostic,
                                             img_size=img_size,
                                             get_time=True)
                results[img_path] = result
                print(f'{img_path}: objects= {len(result["boxes"])}, time= {result["elapsed_time"]}')
                boxes = result['boxes']
                if len(result['boxes']):
                    if res_img_dir:
                        res_path = os.path.join(res_img_dir, item_name)
                        if put_annotations:
                            img = Box.put_box(img, boxes)
                            img = Box.put_text(img, text=[f"{name}_{conf}" for name, conf in
                                                          zip(result['class_names'], result['confidences'])],
                                               org=[(b[0], b[1]) for b in result['boxes']])
                        cv2.imwrite(res_path, img)
                    if res_label_dir:
                        res_path = os.path.join(res_label_dir, split_extension(item_name, '.txt'))
                        xcyc_boxes = Box.box2box(boxes,
                                                 in_format='XYXY',
                                                 to_format='XCYC',
                                                 in_source='Numpy',
                                                 to_source='CV',
                                                 in_relative=False,
                                                 to_relative=True,
                                                 shape_source='Numpy',
                                                 shape=img.shape[:2])
                        with open(res_path, mode='w') as f:
                            for (b1, b2, b3, b4), class_ in zip(xcyc_boxes, result['classes']):
                                f.write(f'{class_} {b1} {b2} {b3} {b4}\n')

        return results



"""Exports a YOLOv5 *.pt model to ONNX and TorchScript formats

Usage:
    $ export PYTHONPATH="$PWD" && python models/export.py --weights yolov5s.pt --img 640 --batch 1
"""

import argparse
import sys
import time

sys.path.append('/')  # to run '$ python *.py' files in subdirectories

import torch
import torch.nn as nn

import models
from models.experimental import attempt_load
from utils.activations import Hardswish, SiLU
from utils.general import colorstr, check_img_size, check_requirements, set_logging
from utils.torch_utils import select_device

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--weights', type=str, default='./yolov5s.pt', help='weights path')  # from yolov5/models/
    parser.add_argument('--img-size', nargs='+', type=int, default=[640, 640], help='image size')  # height, width
    parser.add_argument('--batch-size', type=int, default=1, help='batch size')
    parser.add_argument('--dynamic', action='store_true', help='dynamic ONNX axes')
    parser.add_argument('--grid', action='store_true', help='export Detect() layer grid')
    parser.add_argument('--device', default='cpu', help='cuda device, i.e. 0 or 0,1,2,3 or cpu')
    opt = parser.parse_args()
    opt.img_size *= 2 if len(opt.img_size) == 1 else 1  # expand
    print(opt)
    set_logging()
    t = time.time()

    # Load PyTorch model
    device = select_device(opt.device)
    model = attempt_load(opt.weights, map_location=device)  # load FP32 model
    labels = model.names

    # Checks
    gs = int(max(model.stride))  # grid size (max stride)
    opt.img_size = [check_img_size(x, gs) for x in opt.img_size]  # verify img_size are gs-multiples

    # Input
    img = torch.zeros(opt.batch_size, 3, *opt.img_size).to(device)  # image size(1,3,320,192) iDetection

    # Update model
    for k, m in model.named_modules():
        m._non_persistent_buffers_set = set()  # pytorch 1.6.0 compatibility
        if isinstance(m, models.common.Conv):  # assign export-friendly activations
            if isinstance(m.act, nn.Hardswish):
                m.act = Hardswish()
            elif isinstance(m.act, nn.SiLU):
                m.act = SiLU()
        # elif isinstance(m, models.yolo.Detect):
        #     m.forward = m.forward_export  # assign forward (optional)
    model.model[-1].export = not opt.grid  # set Detect() layer grid export
    y = model(img)  # dry run

    # TorchScript export
    prefix = colorstr('TorchScript:')
    try:
        print(f'\n{prefix} starting export with torch {torch.__version__}...')
        f = opt.weights.replace('.pt', '.torchscript.pt')  # filename
        ts = torch.jit.trace(model, img, strict=False)
        ts.save(f)
        print(f'{prefix} export success, saved as {f}')
    except Exception as e:
        print(f'{prefix} export failure: {e}')

    # ONNX export
    prefix = colorstr('ONNX:')
    try:
        import onnx

        print(f'{prefix} starting export with onnx {onnx.__version__}...')
        f = opt.weights.replace('.pt', '.onnx')  # filename
        torch.onnx.export(model, img, f, verbose=False, opset_version=12, input_names=['images'],
                          output_names=['classes', 'boxes'] if y is None else ['output'],
                          dynamic_axes={'images': {0: 'batch', 2: 'height', 3: 'width'},  # size(1,3,640,640)
                                        'output': {0: 'batch', 2: 'y', 3: 'x'}} if opt.dynamic else None)

        # Checks
        model_onnx = onnx.load(f)  # load onnx model
        onnx.checker.check_model(model_onnx)  # check onnx model
        # print(onnx.helper.printable_graph(model_onnx.graph))  # print

        # Simplify
        try:
            check_requirements(['onnx-simplifier'])
            import onnxsim

            print(f'{prefix} simplifying with onnx-simplifier {onnxsim.__version__}...')
            model_onnx, check = onnxsim.simplify(model_onnx)
            assert check, 'assert check failed'
            onnx.save(model_onnx, f)
        except Exception as e:
            print(f'{prefix} simplifier failure: {e}')
        print(f'{prefix} export success, saved as {f}')
    except Exception as e:
        print(f'{prefix} export failure: {e}')

    # CoreML export
    prefix = colorstr('CoreML:')
    try:
        import coremltools as ct

        print(f'{prefix} starting export with coremltools {onnx.__version__}...')
        # convert model from torchscript and apply pixel scaling as per detect.py
        model = ct.convert(ts, inputs=[ct.ImageType(name='image', shape=img.shape, scale=1 / 255.0, bias=[0, 0, 0])])
        f = opt.weights.replace('.pt', '.mlmodel')  # filename
        model.save(f)
        print(f'{prefix} export success, saved as {f}')
    except Exception as e:
        print(f'{prefix} export failure: {e}')

    # Finish
    print(f'\nExport complete ({time.time() - t:.2f}s). Visualize with https://github.com/lutzroeder/netron.')



# YOLOv5 common modules

import math
from copy import copy
from pathlib import Path

import numpy as np
import pandas as pd
import requests
import torch
import torch.nn as nn
from PIL import Image
from torch.cuda import amp

from utils.datasets import letterbox
from utils.general import non_max_suppression, make_divisible, scale_coords, increment_path, xyxy2xywh
from utils.plots import color_list, plot_one_box
from utils.torch_utils import time_synchronized


def autopad(k, p=None):  # kernel, padding
    # Pad to 'same'
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p


def DWConv(c1, c2, k=1, s=1, act=True):
    # Depthwise convolution
    return Conv(c1, c2, k, s, g=math.gcd(c1, c2), act=act)


class Conv(nn.Module):
    # Standard convolution
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super(Conv, self).__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        self.act = nn.SiLU() if act is True else (act if isinstance(act, nn.Module) else nn.Identity())

    def forward(self, x):
        return self.act(self.bn(self.conv(x)))

    def fuseforward(self, x):
        return self.act(self.conv(x))


class TransformerLayer(nn.Module):
    # Transformer layer https://arxiv.org/abs/2010.11929 (LayerNorm layers removed for better performance)
    def __init__(self, c, num_heads):
        super().__init__()
        self.q = nn.Linear(c, c, bias=False)
        self.k = nn.Linear(c, c, bias=False)
        self.v = nn.Linear(c, c, bias=False)
        self.ma = nn.MultiheadAttention(embed_dim=c, num_heads=num_heads)
        self.fc1 = nn.Linear(c, c, bias=False)
        self.fc2 = nn.Linear(c, c, bias=False)

    def forward(self, x):
        x = self.ma(self.q(x), self.k(x), self.v(x))[0] + x
        x = self.fc2(self.fc1(x)) + x
        return x


class TransformerBlock(nn.Module):
    # Vision Transformer https://arxiv.org/abs/2010.11929
    def __init__(self, c1, c2, num_heads, num_layers):
        super().__init__()
        self.conv = None
        if c1 != c2:
            self.conv = Conv(c1, c2)
        self.linear = nn.Linear(c2, c2)  # learnable position embedding
        self.tr = nn.Sequential(*[TransformerLayer(c2, num_heads) for _ in range(num_layers)])
        self.c2 = c2

    def forward(self, x):
        if self.conv is not None:
            x = self.conv(x)
        b, _, w, h = x.shape
        p = x.flatten(2)
        p = p.unsqueeze(0)
        p = p.transpose(0, 3)
        p = p.squeeze(3)
        e = self.linear(p)
        x = p + e

        x = self.tr(x)
        x = x.unsqueeze(3)
        x = x.transpose(0, 3)
        x = x.reshape(b, self.c2, w, h)
        return x


class Bottleneck(nn.Module):
    # Standard bottleneck
    def __init__(self, c1, c2, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, shortcut, groups, expansion
        super(Bottleneck, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_, c2, 3, 1, g=g)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class BottleneckCSP(nn.Module):
    # CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super(BottleneckCSP, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = nn.Conv2d(c1, c_, 1, 1, bias=False)
        self.cv3 = nn.Conv2d(c_, c_, 1, 1, bias=False)
        self.cv4 = Conv(2 * c_, c2, 1, 1)
        self.bn = nn.BatchNorm2d(2 * c_)  # applied to cat(cv2, cv3)
        self.act = nn.LeakyReLU(0.1, inplace=True)
        self.m = nn.Sequential(*[Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)])

    def forward(self, x):
        y1 = self.cv3(self.m(self.cv1(x)))
        y2 = self.cv2(x)
        return self.cv4(self.act(self.bn(torch.cat((y1, y2), dim=1))))


class C3(nn.Module):
    # CSP Bottleneck with 3 convolutions
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super(C3, self).__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c1, c_, 1, 1)
        self.cv3 = Conv(2 * c_, c2, 1)  # act=FReLU(c2)
        self.m = nn.Sequential(*[Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)])
        # self.m = nn.Sequential(*[CrossConv(c_, c_, 3, 1, g, 1.0, shortcut) for _ in range(n)])

    def forward(self, x):
        return self.cv3(torch.cat((self.m(self.cv1(x)), self.cv2(x)), dim=1))


class C3TR(C3):
    # C3 module with TransformerBlock()
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):
        super().__init__(c1, c2, n, shortcut, g, e)
        c_ = int(c2 * e)
        self.m = TransformerBlock(c_, c_, 4, n)


class SPP(nn.Module):
    # Spatial pyramid pooling layer used in YOLOv3-SPP
    def __init__(self, c1, c2, k=(5, 9, 13)):
        super(SPP, self).__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * (len(k) + 1), c2, 1, 1)
        self.m = nn.ModuleList([nn.MaxPool2d(kernel_size=x, stride=1, padding=x // 2) for x in k])

    def forward(self, x):
        x = self.cv1(x)
        return self.cv2(torch.cat([x] + [m(x) for m in self.m], 1))


class Focus(nn.Module):
    # Focus wh information into c-space
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super(Focus, self).__init__()
        self.conv = Conv(c1 * 4, c2, k, s, p, g, act)
        # self.contract = Contract(gain=2)

    def forward(self, x):  # x(b,c,w,h) -> y(b,4c,w/2,h/2)
        return self.conv(torch.cat([x[..., ::2, ::2], x[..., 1::2, ::2], x[..., ::2, 1::2], x[..., 1::2, 1::2]], 1))
        # return self.conv(self.contract(x))


class Contract(nn.Module):
    # Contract width-height into channels, i.e. x(1,64,80,80) to x(1,256,40,40)
    def __init__(self, gain=2):
        super().__init__()
        self.gain = gain

    def forward(self, x):
        N, C, H, W = x.size()  # assert (H / s == 0) and (W / s == 0), 'Indivisible gain'
        s = self.gain
        x = x.view(N, C, H // s, s, W // s, s)  # x(1,64,40,2,40,2)
        x = x.permute(0, 3, 5, 1, 2, 4).contiguous()  # x(1,2,2,64,40,40)
        return x.view(N, C * s * s, H // s, W // s)  # x(1,256,40,40)


class Expand(nn.Module):
    # Expand channels into width-height, i.e. x(1,64,80,80) to x(1,16,160,160)
    def __init__(self, gain=2):
        super().__init__()
        self.gain = gain

    def forward(self, x):
        N, C, H, W = x.size()  # assert C / s ** 2 == 0, 'Indivisible gain'
        s = self.gain
        x = x.view(N, s, s, C // s ** 2, H, W)  # x(1,2,2,16,80,80)
        x = x.permute(0, 3, 4, 1, 5, 2).contiguous()  # x(1,16,80,2,80,2)
        return x.view(N, C // s ** 2, H * s, W * s)  # x(1,16,160,160)


class Concat(nn.Module):
    # Concatenate a list of tensors along dimension
    def __init__(self, dimension=1):
        super(Concat, self).__init__()
        self.d = dimension

    def forward(self, x):
        return torch.cat(x, self.d)


class NMS(nn.Module):
    # Non-Maximum Suppression (NMS) module
    conf = 0.25  # confidence threshold
    iou = 0.45  # IoU threshold
    classes = None  # (optional list) filter by class

    def __init__(self):
        super(NMS, self).__init__()

    def forward(self, x):
        return non_max_suppression(x[0], conf_thres=self.conf, iou_thres=self.iou, classes=self.classes)


class autoShape(nn.Module):
    # input-robust model wrapper for passing cv2/np/PIL/torch inputs. Includes preprocessing, inference and NMS
    conf = 0.25  # NMS confidence threshold
    iou = 0.45  # NMS IoU threshold
    classes = None  # (optional list) filter by class

    def __init__(self, model):
        super(autoShape, self).__init__()
        self.model = model.eval()

    def autoshape(self):
        print('autoShape already enabled, skipping... ')  # model already converted to model.autoshape()
        return self

    @torch.no_grad()
    def forward(self, imgs, size=640, augment=False, profile=False):
        # Inference from various sources. For height=640, width=1280, RGB images example inputs are:
        #   filename:   imgs = 'data/samples/zidane.jpg'
        #   URI:             = 'https://github.com/ultralytics/yolov5/releases/download/v1.0/zidane.jpg'
        #   OpenCV:          = cv2.imread('image.jpg')[:,:,::-1]  # HWC BGR to RGB x(640,1280,3)
        #   PIL:             = Image.open('image.jpg')  # HWC x(640,1280,3)
        #   numpy:           = np.zeros((640,1280,3))  # HWC
        #   torch:           = torch.zeros(16,3,320,640)  # BCHW (scaled to size=640, 0-1 values)
        #   multiple:        = [Image.open('image1.jpg'), Image.open('image2.jpg'), ...]  # list of images

        t = [time_synchronized()]
        p = next(self.model.parameters())  # for device and type
        if isinstance(imgs, torch.Tensor):  # torch
            with amp.autocast(enabled=p.device.type != 'cpu'):
                return self.model(imgs.to(p.device).type_as(p), augment, profile)  # inference

        # Pre-process
        n, imgs = (len(imgs), imgs) if isinstance(imgs, list) else (1, [imgs])  # number of images, list of images
        shape0, shape1, files = [], [], []  # image and inference shapes, filenames
        for i, im in enumerate(imgs):
            f = f'image{i}'  # filename
            if isinstance(im, str):  # filename or uri
                im, f = np.asarray(Image.open(requests.get(im, stream=True).raw if im.startswith('http') else im)), im
            elif isinstance(im, Image.Image):  # PIL Image
                im, f = np.asarray(im), getattr(im, 'filename', f) or f
            files.append(Path(f).with_suffix('.jpg').name)
            if im.shape[0] < 5:  # image in CHW
                im = im.transpose((1, 2, 0))  # reverse dataloader .transpose(2, 0, 1)
            im = im[:, :, :3] if im.ndim == 3 else np.tile(im[:, :, None], 3)  # enforce 3ch input
            s = im.shape[:2]  # HWC
            shape0.append(s)  # image shape
            g = (size / max(s))  # gain
            shape1.append([y * g for y in s])
            imgs[i] = im  # update
        shape1 = [make_divisible(x, int(self.stride.max())) for x in np.stack(shape1, 0).max(0)]  # inference shape
        x = [letterbox(im, new_shape=shape1, auto=False)[0] for im in imgs]  # pad
        x = np.stack(x, 0) if n > 1 else x[0][None]  # stack
        x = np.ascontiguousarray(x.transpose((0, 3, 1, 2)))  # BHWC to BCHW
        x = torch.from_numpy(x).to(p.device).type_as(p) / 255.  # uint8 to fp16/32
        t.append(time_synchronized())

        with amp.autocast(enabled=p.device.type != 'cpu'):
            # Inference
            y = self.model(x, augment, profile)[0]  # forward
            t.append(time_synchronized())

            # Post-process
            y = non_max_suppression(y, conf_thres=self.conf, iou_thres=self.iou, classes=self.classes)  # NMS
            for i in range(n):
                scale_coords(shape1, y[i][:, :4], shape0[i])

            t.append(time_synchronized())
            return Detections(imgs, y, files, t, self.names, x.shape)


class Detections:
    # detections class for YOLOv5 inference results
    def __init__(self, imgs, pred, files, times=None, names=None, shape=None):
        super(Detections, self).__init__()
        d = pred[0].device  # device
        gn = [torch.tensor([*[im.shape[i] for i in [1, 0, 1, 0]], 1., 1.], device=d) for im in imgs]  # normalizations
        self.imgs = imgs  # list of images as numpy arrays
        self.pred = pred  # list of tensors pred[0] = (xyxy, conf, cls)
        self.names = names  # class names
        self.files = files  # image filenames
        self.xyxy = pred  # xyxy pixels
        self.xywh = [xyxy2xywh(x) for x in pred]  # xywh pixels
        self.xyxyn = [x / g for x, g in zip(self.xyxy, gn)]  # xyxy normalized
        self.xywhn = [x / g for x, g in zip(self.xywh, gn)]  # xywh normalized
        self.n = len(self.pred)  # number of images (batch size)
        self.t = tuple((times[i + 1] - times[i]) * 1000 / self.n for i in range(3))  # timestamps (ms)
        self.s = shape  # inference BCHW shape

    def display(self, pprint=False, show=False, save=False, render=False, save_dir=''):
        colors = color_list()
        for i, (img, pred) in enumerate(zip(self.imgs, self.pred)):
            str = f'image {i + 1}/{len(self.pred)}: {img.shape[0]}x{img.shape[1]} '
            if pred is not None:
                for c in pred[:, -1].unique():
                    n = (pred[:, -1] == c).sum()  # detections per class
                    str += f"{n} {self.names[int(c)]}{'s' * (n > 1)}, "  # add to string
                if show or save or render:
                    for *box, conf, cls in pred:  # xyxy, confidence, class
                        label = f'{self.names[int(cls)]} {conf:.2f}'
                        plot_one_box(box, img, label=label, color=colors[int(cls) % 10])
            img = Image.fromarray(img.astype(np.uint8)) if isinstance(img, np.ndarray) else img  # from np
            if pprint:
                print(str.rstrip(', '))
            if show:
                img.show(self.files[i])  # show
            if save:
                f = self.files[i]
                img.save(Path(save_dir) / f)  # save
                print(f"{'Saved' * (i == 0)} {f}", end=',' if i < self.n - 1 else f' to {save_dir}\n')
            if render:
                self.imgs[i] = np.asarray(img)

    def print(self):
        self.display(pprint=True)  # print results
        print(f'Speed: %.1fms pre-process, %.1fms inference, %.1fms NMS per image at shape {tuple(self.s)}' % self.t)

    def show(self):
        self.display(show=True)  # show results

    def save(self, save_dir='runs/hub/exp'):
        save_dir = increment_path(save_dir, exist_ok=save_dir != 'runs/hub/exp')  # increment save_dir
        Path(save_dir).mkdir(parents=True, exist_ok=True)
        self.display(save=True, save_dir=save_dir)  # save results

    def render(self):
        self.display(render=True)  # render results
        return self.imgs

    def pandas(self):
        # return detections as pandas DataFrames, i.e. print(results.pandas().xyxy[0])
        new = copy(self)  # return copy
        ca = 'xmin', 'ymin', 'xmax', 'ymax', 'confidence', 'class', 'name'  # xyxy columns
        cb = 'xcenter', 'ycenter', 'width', 'height', 'confidence', 'class', 'name'  # xywh columns
        for k, c in zip(['xyxy', 'xyxyn', 'xywh', 'xywhn'], [ca, ca, cb, cb]):
            a = [[x[:5] + [int(x[5]), self.names[int(x[5])]] for x in x.tolist()] for x in getattr(self, k)]  # update
            setattr(new, k, [pd.DataFrame(x, columns=c) for x in a])
        return new

    def tolist(self):
        # return a list of Detections objects, i.e. 'for result in results.tolist():'
        x = [Detections([self.imgs[i]], [self.pred[i]], self.names, self.s) for i in range(self.n)]
        for d in x:
            for k in ['imgs', 'pred', 'xyxy', 'xyxyn', 'xywh', 'xywhn']:
                setattr(d, k, getattr(d, k)[0])  # pop out of list
        return x

    def __len__(self):
        return self.n


class Classify(nn.Module):
    # Classification head, i.e. x(b,c1,20,20) to x(b,c2)
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1):  # ch_in, ch_out, kernel, stride, padding, groups
        super(Classify, self).__init__()
        self.aap = nn.AdaptiveAvgPool2d(1)  # to x(b,c1,1,1)
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g)  # to x(b,c2,1,1)
        self.flat = nn.Flatten()

    def forward(self, x):
        z = torch.cat([self.aap(y) for y in (x if isinstance(x, list) else [x])], 1)  # cat if list
        return self.flat(self.conv(z))  # flatten to x(b,c2)



import os

IMG_EXTENSIONS = [
    '.jpg', '.JPG', '.jpeg', '.JPEG',
    '.png', '.PNG', '.ppm', '.PPM', '.bmp', '.BMP', '.tiff'
]


def is_image_file(filename):
    return any(filename.endswith(extension) for extension in IMG_EXTENSIONS)


def split_extension(path, extension=None, suffix=None):
    remain, extension_ = os.path.splitext(path)
    if extension and suffix:
        return remain + suffix + extension
    elif extension is None and suffix:
        return remain + suffix + extension_
    elif extension:
        return remain + extension
    return remain, extension_


def split_all(path):
    all_parts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            all_parts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            all_parts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            all_parts.insert(0, parts[1])
    return all_parts



import math
from collections import defaultdict
import numpy as np
from .box_utils import nms, _preprocess
from deep_utils.utils.resize_utils.main_resize import resize


def run_first_stage(image, net, scale, threshold, device):
    """Run P-Net, generate bounding boxes, and do NMS.
from deep_utils.utils

    Arguments:
        image: an instance of PIL.Image.
        net: an instance of pytorch's nn.Module, P-Net.
        scale: a float number,
            scale width and height of the image by this number.
        threshold: a float number,
            threshold on the probability of a face when generating
            bounding boxes from predictions of the net.

    Returns:
        a float numpy array of shape [n_boxes, 9],
            bounding boxes with scores and offsets (4 + 1 + 4).
    """

    # scale the image and convert it to a float array
    width, height = image.shape[1:3]
    sw, sh = math.ceil(width * scale), math.ceil(height * scale)
    img = resize(image, (sw, sh))
    img = img.transpose((0, 2, 1, 3))
    img = (img - 127.5) * 0.0078125
    output = net.predict(img)
    out0 = output[0].transpose((0, 3, 2, 1))
    out1 = output[1].transpose((0, 3, 2, 1))
    bboxes = []
    for img_n in range(img.shape[0]):
        probs = out0[img_n, 1, :, :]
        offsets = out1[img_n]
        # probs: probability of a face at each sliding window
        # offsets: transformations to true bounding boxes

        boxes = _generate_bboxes(probs, offsets, scale, threshold)
        if len(boxes) == 0:
            return None

        keep = nms(boxes[:, 0:5], overlap_threshold=0.5)
        bboxes.append(boxes[keep])
    return bboxes


def _generate_bboxes(probs, offsets, scale, threshold):
    """Generate bounding boxes at places
    where there is probably a face.

    Arguments:
        probs: a float numpy array of shape [n, m].
        offsets: a float numpy array of shape [1, 4, n, m].
        scale: a float number,
            width and height of the image were scaled by this number.
        threshold: a float number.

    Returns:
        a float numpy array of shape [n_boxes, 9]
    """
    # applying P-Net is equivalent, in some sense, to
    # moving 12x12 window with stride 2
    stride = 2
    cell_size = 12

    # indices of boxes where there is probably a face
    inds = np.where(probs > threshold)

    if inds[0].size == 0:
        return np.array([])

    # transformations of bounding boxes
    tx1, ty1, tx2, ty2 = [offsets[i, inds[0], inds[1]] for i in range(4)]
    # they are defined as:
    # w = x2 - x1 + 1
    # h = y2 - y1 + 1
    # x1_true = x1 + tx1*w
    # x2_true = x2 + tx2*w
    # y1_true = y1 + ty1*h
    # y2_true = y2 + ty2*h

    offsets = np.array([tx1, ty1, tx2, ty2])
    score = probs[inds[0], inds[1]]

    # P-Net is applied to scaled images
    # so we need to rescale bounding boxes back
    bounding_boxes = np.vstack([
        np.round((stride * inds[1] + 1.0) / scale),
        np.round((stride * inds[0] + 1.0) / scale),
        np.round((stride * inds[1] + 1.0 + cell_size) / scale),
        np.round((stride * inds[0] + 1.0 + cell_size) / scale),
        score, offsets
    ])
    return bounding_boxes.T



def ONet(weight_path='onet.h5'):
    from tensorflow.keras.layers import Conv2D, Input, MaxPool2D, Flatten, Dense, Permute, PReLU
    from tensorflow.keras.models import Model
    input = Input(shape=[48, 48, 3])
    x = Conv2D(32, (3, 3), strides=1, padding='valid', name='conv1')(input)
    x = PReLU(shared_axes=[1, 2], name='prelu1')(x)
    x = MaxPool2D(pool_size=3, strides=2, padding='same')(x)
    x = Conv2D(64, (3, 3), strides=1, padding='valid', name='conv2')(x)
    x = PReLU(shared_axes=[1, 2], name='prelu2')(x)
    x = MaxPool2D(pool_size=3, strides=2)(x)
    x = Conv2D(64, (3, 3), strides=1, padding='valid', name='conv3')(x)
    x = PReLU(shared_axes=[1, 2], name='prelu3')(x)
    x = MaxPool2D(pool_size=2)(x)
    x = Conv2D(128, (2, 2), strides=1, padding='valid', name='conv4')(x)
    x = PReLU(shared_axes=[1, 2], name='prelu4')(x)
    x = Permute((3, 2, 1))(x)
    x = Flatten()(x)
    x = Dense(256, name='conv5')(x)
    x = PReLU(name='prelu5')(x)

    classifier = Dense(2, activation='softmax', name='conv6-1')(x)
    bbox_regress = Dense(4, name='conv6-2')(x)
    landmark_regress = Dense(10, name='conv6-3')(x)
    model = Model([input], [classifier, bbox_regress, landmark_regress])
    model.load_weights(weight_path, by_name=True)

    return model


def RNet(weight_path='rnet.h5'):
    from tensorflow.keras.layers import Conv2D, Input, MaxPool2D, Flatten, Dense, Permute, PReLU
    from tensorflow.keras.models import Model
    input = Input(shape=[24, 24, 3])  # change this shape to [None,None,3] to enable arbitraty shape input
    x = Conv2D(28, (3, 3), strides=1, padding='valid', name='conv1')(input)
    x = PReLU(shared_axes=[1, 2], name='prelu1')(x)
    x = MaxPool2D(pool_size=3, strides=2, padding='same')(x)

    x = Conv2D(48, (3, 3), strides=1, padding='valid', name='conv2')(x)
    x = PReLU(shared_axes=[1, 2], name='prelu2')(x)
    x = MaxPool2D(pool_size=3, strides=2)(x)

    x = Conv2D(64, (2, 2), strides=1, padding='valid', name='conv3')(x)
    x = PReLU(shared_axes=[1, 2], name='prelu3')(x)
    x = Permute((3, 2, 1))(x)
    x = Flatten()(x)
    x = Dense(128, name='conv4')(x)
    x = PReLU(name='prelu4')(x)
    classifier = Dense(2, activation='softmax', name='conv5-1')(x)
    bbox_regress = Dense(4, name='conv5-2')(x)
    model = Model([input], [classifier, bbox_regress])
    model.load_weights(weight_path, by_name=True)
    return model


def PNet(weight_path='pnet.h5'):
    from tensorflow.keras.layers import Conv2D, Input, MaxPool2D, Flatten, Dense, Permute, PReLU
    from tensorflow.keras.models import Model
    input = Input(shape=[None, None, 3])
    x = Conv2D(10, (3, 3), strides=1, padding='valid', name='conv1')(input)
    x = PReLU(shared_axes=[1, 2], name='PReLU1')(x)
    x = MaxPool2D(pool_size=2)(x)
    x = Conv2D(16, (3, 3), strides=1, padding='valid', name='conv2')(x)
    x = PReLU(shared_axes=[1, 2], name='PReLU2')(x)
    x = Conv2D(32, (3, 3), strides=1, padding='valid', name='conv3')(x)
    x = PReLU(shared_axes=[1, 2], name='PReLU3')(x)
    classifier = Conv2D(2, (1, 1), activation='softmax', name='conv4-1')(x)
    bbox_regress = Conv2D(4, (1, 1), name='conv4-2')(x)
    model = Model([input], [classifier, bbox_regress])
    model.load_weights(weight_path, by_name=True)
    return model



import sys
import argparse


class frozendict:
    def __init__(self, **kwargs):
        super(frozendict, self).__init__()
        self.repo = dict()
        for key, item in kwargs.items():
            self.repo[key] = item

    def __getitem__(self, item):
        return self.repo[item]

    def __setitem__(self, key, value):
        raise TypeError(
            f"frozendict object does not support updating"
        )


def shift_lst(lst: list, move_forward):
    return lst[-move_forward:] + lst[:-move_forward]


if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
    from deep_utils.utils.utils.dictnamedtuple_38 import dictnamedtuple
else:
    from deep_utils.utils.utils.dictnamedtuple_37 import dictnamedtuple


def easy_argparse(*args):
    parser = argparse.ArgumentParser()

    for arg in args:
        tp = type(arg)
        if tp == dict:
            name = arg.pop('name')
            parser.add_argument(name, **arg)
        else:
            raise TypeError(
                f"easy_argparse just supports dictionaries"
            )
    return parser.parse_args()



import numpy as np
from deep_utils.vision.face_detection.main.main_face_detection import FaceDetector
from deep_utils.utils.lib_utils.lib_decorators import get_from_config, expand_input, get_elapsed_time, rgb2bgr
from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.box_utils.boxes import Box, Point
from .config import Config
from .src.get_nets import PNet, RNet, ONet
from .src.box_utils import nms, calibrate_box, get_image_boxes, convert_to_square
from .src.first_stage import run_first_stage


class MTCNNTFFaceDetector(FaceDetector):

    def __init__(self, **kwargs):
        super().__init__(name=self.__class__.__name__,
                         file_path=__file__,
                         download_variables=("pnet", "onet", 'rnet'),
                         **kwargs)
        self.config: Config

    @download_decorator
    def load_model(self):
        # LOAD MODELS

        pnet = PNet(self.config.pnet)
        rnet = RNet(self.config.rnet)
        onet = ONet(self.config.onet)
        self.model = dict(pnet=pnet, rnet=rnet, onet=onet)

    @rgb2bgr('rgb')
    @get_elapsed_time
    @expand_input(3)
    @get_from_config
    def detect_faces(self,
                     img,
                     is_rgb,
                     min_face_size=None,
                     thresholds=None,
                     nms_thresholds=None,
                     min_detection_size=None,
                     factor=None,
                     confidence=None,
                     get_time=False):

        # BUILD AN IMAGE PYRAMID
        width, height = img.shape[1:3]
        min_length = min(height, width)

        # scales for scaling the image
        scales = []

        # scales the image so that
        # minimum size that we can detect equals to
        # minimum face size that we want to detect
        m = min_detection_size / min_face_size
        min_length *= m

        factor_count = 0
        while min_length > min_detection_size:
            scales.append(m * factor ** factor_count)
            min_length *= factor
            factor_count += 1

        # STAGE 1

        # it will be returned
        bounding_boxes = []

        # run P-Net on different scales
        for s in scales:
            boxes = run_first_stage(img,
                                    self.model["pnet"],
                                    scale=s,
                                    threshold=thresholds[0],
                                    device=self.config.device)
            bounding_boxes.append(boxes)
        bounding_boxes_ = bounding_boxes
        image_boxes = []
        join_bounding_boxes = []
        for img_n in range(img.shape[0]):
            bounding_boxes = [box[img_n] for box in bounding_boxes_ if box is not None]
            # bounding_boxes = [i for i in bounding_boxes if i is not None]
            bounding_boxes = np.vstack(bounding_boxes)

            keep = nms(bounding_boxes[:, 0:5], nms_thresholds[0])
            bounding_boxes = bounding_boxes[keep]

            # use offsets predicted by pnet to transform bounding boxes
            bounding_boxes = calibrate_box(bounding_boxes[:, 0:5], bounding_boxes[:, 5:])
            # shape [n_boxes, 5]

            bounding_boxes = convert_to_square(bounding_boxes)
            bounding_boxes[:, 0:4] = np.round(bounding_boxes[:, 0:4])

            # STAGE 2

            img_boxes = get_image_boxes(bounding_boxes, img[img_n], size=24)
            if img_boxes.size != 0:
                image_boxes.append(img_boxes)
                join_bounding_boxes.append(bounding_boxes)
        if len(image_boxes) == 0:
            return dict(boxes=[], confidences=[], landmarks=[])
        bounding_boxes_ = np.vstack(join_bounding_boxes)
        split = [0]
        for img_box in image_boxes:
            split.append(img_box.shape[0] + split[-1])
        img_boxes = np.concatenate(image_boxes)
        img_boxes = img_boxes.transpose([0, 2, 3, 1])
        output = self.model["rnet"].predict(img_boxes)
        image_boxes = []
        join_bounding_boxes = []
        for img_n in range(img.shape[0]):
            offsets = output[1][split[img_n]: split[img_n + 1]]  # shape [n_boxes, 4]
            probs = output[0][split[img_n]: split[img_n + 1]]  # shape [n_boxes, 2]
            bounding_boxes = bounding_boxes_[split[img_n]: split[img_n + 1]]

            keep = np.where(probs[:, 1] > thresholds[1])[0]
            bounding_boxes = bounding_boxes[keep]
            bounding_boxes[:, 4] = probs[keep, 1].reshape((-1,))
            offsets = offsets[keep]

            keep = nms(bounding_boxes, nms_thresholds[1])
            bounding_boxes = bounding_boxes[keep]
            bounding_boxes = calibrate_box(bounding_boxes, offsets[keep])
            bounding_boxes = convert_to_square(bounding_boxes)
            bounding_boxes[:, 0:4] = np.round(bounding_boxes[:, 0:4])

            # STAGE 3

            img_boxes = get_image_boxes(bounding_boxes, img[img_n], size=48)
            if img_boxes.size != 0:
                image_boxes.append(img_boxes)
                join_bounding_boxes.append(bounding_boxes)
        if len(image_boxes) == 0:
            return dict(boxes=[], confidences=[], landmarks=[])
        split = [0]
        for img_box in image_boxes:
            split.append(img_box.shape[0] + split[-1])
        img_boxes = np.vstack(image_boxes)
        bounding_boxes_ = np.vstack(join_bounding_boxes)
        img_boxes = img_boxes.transpose([0, 2, 3, 1])
        output = self.model["onet"].predict(img_boxes)
        boxes_, confidences_, landmarks_ = [], [], []
        for img_n in range(img.shape[0]):
            bounding_boxes = bounding_boxes_[split[img_n]: split[img_n + 1]]
            landmarks = output[2][split[img_n]: split[img_n + 1]]
            offsets = output[1][split[img_n]: split[img_n + 1]]  # shape [n_boxes, 4]
            probs = output[0][split[img_n]: split[img_n + 1]]  # shape [n_boxes, 2]
            keep = np.where(probs[:, 1] > thresholds[2])[0]
            bounding_boxes = bounding_boxes[keep]
            bounding_boxes[:, 4] = probs[keep, 1].reshape((-1,))
            offsets = offsets[keep]
            landmarks = landmarks[keep]

            width = bounding_boxes[:, 2] - bounding_boxes[:, 0] + 1.0
            height = bounding_boxes[:, 3] - bounding_boxes[:, 1] + 1.0
            xmin, ymin = bounding_boxes[:, 0], bounding_boxes[:, 1]
            landmarks[:, 0:5] = np.expand_dims(xmin, 1) + np.expand_dims(width, 1) * landmarks[:, 0:5]
            landmarks[:, 5:10] = np.expand_dims(ymin, 1) + np.expand_dims(height, 1) * landmarks[:, 5:10]

            bounding_boxes = calibrate_box(bounding_boxes, offsets)
            keep = nms(bounding_boxes, nms_thresholds[2], mode='min')
            landmarks = landmarks[keep]
            boxes, confidences = bounding_boxes[keep][:, :4], bounding_boxes[keep][:, 4]
            keep = confidences >= confidence
            boxes, confidences, landmarks = boxes[keep], confidences[keep], landmarks[keep]
            boxes = Box.box2box(boxes, in_source=Box.BoxSource.CV, to_source=Box.BoxSource.Numpy)
            boxes_.append(boxes)
            confidences_.append(confidences)
            if len(landmarks) != 0:
                landmarks = [[Point.point2point((landmarks[j][i], landmarks[j][5 + i]),
                                                in_source='CV', to_source='Numpy') for i in range(5)] for j in
                             range(landmarks.shape[0])]
            face_points = ["nose", "mouth_right", "right_eye", "left_eye", "mouth_left"]
            for i in range(0, len(landmarks)):
                face_dict = {}
                for points, face in zip(landmarks[i], face_points):
                    face_dict[face] = points
                landmarks_.append(face_dict)
        return dict(boxes=boxes_, confidences=confidences_, landmarks=landmarks_)



from collections.abc import Callable
import importlib
import warnings


def import_module(module_name, things_to_import):
    try:
        new_module = importlib.import_module(module_name)
        return getattr(new_module, things_to_import)
    except ModuleNotFoundError as e:
        warnings.warn(f"\n{e}. If you don't use {things_to_import} ignore this message.", stacklevel=2)


def list_utils(module_dict):
    def list_models():
        detection_models = ""
        for name, _ in module_dict.items():
            detection_models += f'{name}\n'
        return detection_models

    return list_models


def loader(module_dict, list_models) -> Callable:
    def module_loader(name, **kwargs):
        if name not in module_dict:
            raise Exception(f'{name} model is not supported. Supported models are:\n{list_models()}')
        model = module_dict[name](**kwargs)
        return model

    return module_loader



import numpy as np
from deep_utils.main_abs.cv2.cv2_caffe import CV2Caffe
from deep_utils.vision.face_detection.main import FaceDetector
from deep_utils.utils.lib_utils.lib_decorators import get_from_config, expand_input, get_elapsed_time, rgb2bgr
from deep_utils.utils.resize_utils.main_resize import get_img_shape
from deep_utils.utils.box_utils.boxes import Box


class SSDCV2CaffeFaceDetector(FaceDetector):

    def __init__(self, **kwargs):
        super().__init__(name=self.__class__.__name__, file_path=__file__, **kwargs)

    def load_model(self):
        self.model = CV2Caffe(self.config)

    @rgb2bgr('bgr')
    @get_elapsed_time
    @expand_input(3)
    @get_from_config
    def detect_faces(self,
                     img,
                     is_rgb,
                     resize_size=None,
                     img_scale_factor=None,
                     img_mean=None,
                     resize_mode=None,
                     confidence=None,
                     get_time=False):
        faces = self.model.forward(img,
                                   resize_size=resize_size,
                                   img_scale_factor=img_scale_factor,
                                   img_mean=img_mean,
                                   resize_mode=resize_mode
                                   )
        n_image, h, w, _ = get_img_shape(img)
        boxes = [[] for _ in range(n_image)]
        confidences = [[] for _ in range(n_image)]
        for i in range(faces.shape[2]):
            img_number = int(faces[0, 0, i, 0])
            face_confidence = faces[0, 0, i, 2]
            if face_confidence < confidence:
                continue
            box = faces[0, 0, i, 3:7] * np.array([w, h, w, h])
            box = Box.box2box(box, in_source='CV', to_source='Numpy', return_int=True)
            boxes[img_number].append(box)
            confidences[img_number].append(face_confidence)
        return dict(boxes=boxes, confidences=confidences)


if __name__ == '__main__':
    model = SSDCV2CaffeFaceDetector()



from deep_utils.vision.face_detection.main import FaceDetector
from deep_utils.utils.lib_utils.lib_decorators import get_from_config, expand_input, get_elapsed_time, rgb2bgr
from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.box_utils.boxes import Box, Point
from .config import Config


class HaarcascadeCV2FaceDetector(FaceDetector):

    def __init__(self, **kwargs):
        super().__init__(name=self.__class__.__name__,
                         file_path=__file__,
                         download_variables=("haarcascade_frontalface",
                                             "haarcascade_eye",
                                             "haarcascade_nose",
                                             "landmarks"),
                         **kwargs)
        self.config: Config

    @download_decorator
    def load_model(self):
        import cv2
        face_detector = cv2.CascadeClassifier(self.config.haarcascade_frontalface)
        nose_detector = cv2.CascadeClassifier(self.config.haarcascade_nose)
        eye_detector = cv2.CascadeClassifier(self.config.haarcascade_eye)
        landmark_detector = cv2.face.createFacemarkLBF()
        landmark_detector.loadModel(self.config.landmarks)
        self.model = dict(face=face_detector, nose=nose_detector, eye=eye_detector, landmarks=landmark_detector)

    @get_elapsed_time
    @expand_input(3)
    @get_from_config
    @rgb2bgr('gray')
    def detect_faces(self,
                     img,
                     is_rgb,
                     scaleFactor=None,
                     minNeighbors=None,
                     minSize=None,
                     maxSize=None,
                     flags=None,
                     get_landmarks=True,
                     get_nose=True,
                     get_eye=True,
                     get_time=False,
                     ):
        boxes, landmarks_, confidences, eye_poses, nose_poses = [], [], [], [], []
        for image in img:
            faces = self.model['face'].detectMultiScale(image,
                                                        scaleFactor=scaleFactor,
                                                        minNeighbors=minNeighbors,
                                                        minSize=minSize,
                                                        maxSize=maxSize,
                                                        flags=flags)
            boxes.append(Box.box2box(faces, in_source='CV', to_source='Numpy', in_format='XYWH', to_format='XYXY'))
            if get_nose:
                nose_pos = self.model['nose'].detectMultiScale(image,
                                                               scaleFactor=scaleFactor,
                                                               minNeighbors=minNeighbors,
                                                               minSize=minSize,
                                                               maxSize=maxSize,
                                                               flags=flags)
                nose_pos = Point.point2point(nose_pos, in_source='CV', to_source='Numpy')
                nose_poses.append(nose_pos)
            if get_eye:
                eye_pos = self.model['eye'].detectMultiScale(image,
                                                             scaleFactor=scaleFactor,
                                                             minNeighbors=minNeighbors,
                                                             minSize=minSize,
                                                             maxSize=maxSize,
                                                             flags=flags)
                eye_pos = Point.point2point(eye_pos, in_source='CV', to_source='Numpy')
                eye_poses.append(eye_pos)
            if len(faces) != 0 and get_landmarks:
                _, landmarks = self.model['landmarks'].fit(image, faces)
                landmarks = [Point.point2point(face_landmarks[0].tolist(),
                                               in_source='CV', to_source='Numpy') for face_landmarks in landmarks]
                landmarks_.append(landmarks)
        return dict(boxes=boxes,
                    confidences=confidences,
                    landmarks=landmarks_,
                    eye_poses=eye_poses,
                    nose_poses=nose_poses)



class Config:
    def __init__(self):
        self.device = 'cpu'
        self.min_face_size = 20.0
        self.thresholds = (0.6, 0.7, 0.8)
        self.nms_thresholds = (0.7, 0.7, 0.7)
        self.min_detection_size = 12
        self.factor = 0.707
        self.confidence = 0.8
        self.pnet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/pnet.npy"
        self.pnet_cache = 'weights/vision/face_detection/mtcnn/torch/pnet.npy'
        self.pnet = None
        self.rnet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/rnet.npy"
        self.rnet_cache = 'weights/vision/face_detection/mtcnn/torch/rnet.npy'
        self.rnet = None
        self.onet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/onet.npy"
        self.onet_cache = 'weights/vision/face_detection/mtcnn/torch/onet.npy'
        self.onet = None



import sys
import os
import setuptools
from setuptools.command.install import install

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = "0.7.0"


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setuptools.setup(
    name="deep_utils",
    version=VERSION,
    author="Pooya Mohammadi Kazaj",
    author_email="pooyamohammadikazaj@gmial.com",
    download_url=f"https://github.com/Practical-AI/deep_utils/archive/refs/tags/{VERSION}.tar.gz",
    description="Deep Utils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy', "requests", "tqdm"
    ],
    cmdclass={
        'verify': VerifyVersionCommand,
    }
)



from typing import Union


class CV2Config:
    def __init__(self):
        self.prototxt = None
        self.caffemodel = None
        self.resize_mode = 'cv2'
        self.resize_size = (300, 300)
        self.img_mean: Union[tuple, None] = None
        self.img_scale_factor = 1



def unzip(zip_path, zip_dir='.'):
    from zipfile import ZipFile
    with ZipFile(zip_path, 'r') as zip:
        print(f'extracting {zip_path}')
        zip.extractall(zip_dir)
        print(f'extracting is done!')



from typing import Union
from abc import abstractmethod
import numpy as np
from deep_utils.main_abs.main import MainClass


class ObjectDetector(MainClass):
    def __init__(self, name, file_path, *args, **kwargs):
        super().__init__(name, file_path=file_path, **kwargs)
        self.confidences: Union[list, None, np.ndarray] = None
        self.boxes: Union[list, None, np.ndarray] = None
        self.classes: Union[list, None, np.ndarray] = None
        self.class_names: Union[list, None, np.ndarray] = None

    @abstractmethod
    def detect_objects(self, img, is_rgb, confidence=None, iou_thresh=None, classes=None, get_time=False, **kwargs):
        pass

    @abstractmethod
    def detect_dir(self, dir_, confidence=None, iou_thresh=None, classes=None,
                   extensions=('.png', '.jpg', '.jpeg'), save_results=False, save_in_file=False, **kwargs):
        pass



class Config:
    def __init__(self):
        self.device = 'cpu'
        self.min_face_size = 20.0
        self.thresholds = (0.6, 0.7, 0.8)
        self.nms_thresholds = (0.7, 0.7, 0.7)
        self.min_detection_size = 12
        self.factor = 0.707
        self.confidence = 0.8
        self.pnet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/pnet.h5"
        self.pnet_cache = 'weights/vision/face_detection/mtcnn/tf/pnet.h5'
        self.pnet = None
        self.rnet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/rnet.h5"
        self.rnet_cache = 'weights/vision/face_detection/mtcnn/tf/rnet.h5'
        self.rnet = None
        self.onet_url = "https://github.com/Practical-AI/deep_utils/releases/download/0.1.0/onet.h5"
        self.onet_cache = 'weights/vision/face_detection/mtcnn/tf/onet.h5'
        self.onet = None



from deep_utils.main_abs.cv2.cv2_main import CV2Main
from deep_utils.utils.lib_utils.download_utils import download_decorator
from .cv2_config import CV2Config


class CV2Caffe(CV2Main):
    def __init__(self, config: CV2Config):
        super().__init__(config=config)
        self.download_variables = ('prototxt', 'caffemodel')
        self.load_model()

    @download_decorator
    def load_model(self):
        import cv2
        self.model = cv2.dnn.readNetFromCaffe(
            self.config.prototxt,
            self.config.caffemodel
        )




import torch
import torch.nn as nn
import torch.nn.functional as F
from ...utils.box_utils import match, log_sum_exp


class MultiBoxLoss(nn.Module):
    """SSD Weighted Loss Function
    Compute Targets:
        1) Produce Confidence Target Indices by matching  ground truth boxes
           with (default) 'priorboxes' that have jaccard index > threshold parameter
           (default threshold: 0.5).
        2) Produce localization target by 'encoding' variance into offsets of ground
           truth boxes and their matched  'priorboxes'.
        3) Hard negative mining to filter the excessive number of negative examples
           that comes with using a large number of default bounding boxes.
           (default negative:positive ratio 3:1)
    Objective Loss:
        L(x,c,l,g) = (Lconf(x, c) + αLloc(x,l,g)) / N
        Where, Lconf is the CrossEntropy Loss and Lloc is the SmoothL1 Loss
        weighted by α which is set to 1 by cross val.
        Args:
            c: class confidences,
            l: predicted boxes,
            g: ground truth boxes
            N: number of matched default boxes
        See: https://arxiv.org/pdf/1512.02325.pdf for more details.
    """

    def __init__(self, num_classes, overlap_thresh, prior_for_matching, bkg_label, neg_mining, neg_pos, neg_overlap,
                 encode_target):
        super(MultiBoxLoss, self).__init__()
        self.num_classes = num_classes
        self.threshold = overlap_thresh
        self.background_label = bkg_label
        self.encode_target = encode_target
        self.use_prior_for_matching = prior_for_matching
        self.do_neg_mining = neg_mining
        self.negpos_ratio = neg_pos
        self.neg_overlap = neg_overlap
        self.variance = [0.1, 0.2]

    def forward(self, predictions, priors, targets):
        """Multibox Loss
        Args:
            predictions (tuple): A tuple containing loc preds, conf preds,
            and prior boxes from SSD net.
                conf shape: torch.size(batch_size,num_priors,num_classes)
                loc shape: torch.size(batch_size,num_priors,4)
                priors shape: torch.size(num_priors,4)

            ground_truth (tensor): Ground truth boxes and labels for a batch,
                shape: [batch_size,num_objs,5] (last idx is the label).
        """

        loc_data, conf_data, landm_data = predictions
        priors = priors
        num = loc_data.size(0)
        num_priors = (priors.size(0))

        # match priors (default boxes) and ground truth boxes
        loc_t = torch.Tensor(num, num_priors, 4)
        landm_t = torch.Tensor(num, num_priors, 10)
        conf_t = torch.LongTensor(num, num_priors)
        for idx in range(num):
            truths = targets[idx][:, :4].data
            labels = targets[idx][:, -1].data
            landms = targets[idx][:, 4:14].data
            defaults = priors.data
            match(self.threshold, truths, defaults, self.variance, labels, landms, loc_t, conf_t, landm_t, idx)

        zeros = torch.tensor(0).cuda()
        # landm Loss (Smooth L1)
        # Shape: [batch,num_priors,10]
        pos1 = conf_t > zeros
        num_pos_landm = pos1.long().sum(1, keepdim=True)
        N1 = max(num_pos_landm.data.sum().float(), 1)
        pos_idx1 = pos1.unsqueeze(pos1.dim()).expand_as(landm_data)
        landm_p = landm_data[pos_idx1].view(-1, 10)
        landm_t = landm_t[pos_idx1].view(-1, 10)
        loss_landm = F.smooth_l1_loss(landm_p, landm_t, reduction='sum')

        pos = conf_t != zeros
        conf_t[pos] = 1

        # Localization Loss (Smooth L1)
        # Shape: [batch,num_priors,4]
        pos_idx = pos.unsqueeze(pos.dim()).expand_as(loc_data)
        loc_p = loc_data[pos_idx].view(-1, 4)
        loc_t = loc_t[pos_idx].view(-1, 4)
        loss_l = F.smooth_l1_loss(loc_p, loc_t, reduction='sum')

        # Compute max conf across batch for hard negative mining
        batch_conf = conf_data.view(-1, self.num_classes)
        loss_c = log_sum_exp(batch_conf) - batch_conf.gather(1, conf_t.view(-1, 1))

        # Hard Negative Mining
        loss_c[pos.view(-1, 1)] = 0  # filter out pos boxes for now
        loss_c = loss_c.view(num, -1)
        _, loss_idx = loss_c.sort(1, descending=True)
        _, idx_rank = loss_idx.sort(1)
        num_pos = pos.long().sum(1, keepdim=True)
        num_neg = torch.clamp(self.negpos_ratio * num_pos, max=pos.size(1) - 1)
        neg = idx_rank < num_neg.expand_as(idx_rank)

        # Confidence Loss Including Positive and Negative Examples
        pos_idx = pos.unsqueeze(2).expand_as(conf_data)
        neg_idx = neg.unsqueeze(2).expand_as(conf_data)
        conf_p = conf_data[(pos_idx + neg_idx).gt(0)].view(-1, self.num_classes)
        targets_weighted = conf_t[(pos + neg).gt(0)]
        loss_c = F.cross_entropy(conf_p, targets_weighted, reduction='sum')

        # Sum of losses: L(x,c,l,g) = (Lconf(x, c) + αLloc(x,l,g)) / N
        N = max(num_pos.data.sum().float(), 1)
        loss_l /= N
        loss_c /= N
        loss_landm /= N1

        return loss_l, loss_c, loss_landm



import numpy as np
from deep_utils.utils.resize_utils.main_resize import resize
from deep_utils.utils.box_utils.boxes import Box


def group_show(images, size=(128, 128), n_channels=3, texts=None, text_org=None, text_kwargs=None, title=None,
               title_org=None, title_kwargs=None):
    import math
    title_kwargs = dict() if title_kwargs is None else title_kwargs
    text_kwargs = dict() if text_kwargs is None else text_kwargs
    n_images = len(images)
    n = math.ceil(math.sqrt(len(images)))
    columns = n
    rows = math.ceil(n_images / columns)
    img_size = (rows * size[0], columns * size[1])
    img = np.zeros((img_size[0], img_size[1], n_channels), dtype=np.uint8)
    i = 0
    for r in range(rows):
        for c in range(columns):
            resized_img = resize(images[i], size)
            if texts is not None:
                org = (size[1] // 10, size[0] // 2 - size[0] // 5) if text_org is None else text_org[i]
                resized_img = Box.put_text(resized_img, texts[i], org, **text_kwargs)
            img[r * size[0]:(r + 1) * size[0], c * size[1]:(c + 1) * size[1]] = resized_img
            i += 1
            if i == len(images):
                break
        if i == len(images):
            break
    img = np.concatenate([np.zeros((size[0] // 5, img_size[1], 3)), img])
    if title:
        if title_org is None:
            title_org = (size[1] // 10, img_size[0] // 2 - img_size[0] // 5) if text_org is None else text_org[i]
        img = Box.put_text(img, text=title, org=title_org, **title_kwargs)
    img = img.astype(np.uint8)
    return img



from .lib_utils import *
from .os_utils import *
from .box_utils import *
from .resize_utils import *
from .opencv import *
from .dir_utils import *
from .matplotlib_utils import *
from .utils import *
from .imshow_utils import *
